#!/usr/bin/python
# Copyright (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ipa_service
author: Cédric Parent (@cprh)
short_description: Manage FreeIPA service
description:
  - Add and delete an IPA service using IPA API.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  krbcanonicalname:
    description:
      - Principal of the service.
      - Can not be changed as it is the unique identifier.
    required: true
    aliases: ["name"]
    type: str
  hosts:
    description:
      - Defines the list of C(ManagedBy) hosts.
    type: list
    elements: str
  force:
    description:
      - Force principal name even if host is not in DNS.
    type: bool
  skip_host_check:
    description:
      - Force service to be created even when host object does not exist to manage it.
      - This is only used on creation, not for updating existing services.
    type: bool
    default: false
    version_added: 4.7.0
  state:
    description: State to ensure.
    default: present
    choices: ["absent", "present"]
    type: str
extends_documentation_fragment:
  - community.general._ipa.documentation
  - community.general._ipa.connection_notes
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Ensure service is present
  community.general.ipa_service:
    name: http/host01.example.com
    state: present
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Ensure service is absent
  community.general.ipa_service:
    name: http/host01.example.com
    state: absent
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Changing Managing hosts list
  community.general.ipa_service:
    name: http/host01.example.com
    hosts:
      - host01.example.com
      - host02.example.com
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret
"""

RETURN = r"""
service:
  description: Service as returned by IPA API.
  returned: always
  type: dict
"""

import traceback

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ipa import IPAClient, ipa_argument_spec


class ServiceIPAClient(IPAClient):
    def __init__(self, module, host, port, protocol):
        super().__init__(module, host, port, protocol)

    def service_find(self, name):
        return self._post_json(method="service_find", name=None, item={"all": True, "krbcanonicalname": name})

    def service_add(self, name, service):
        return self._post_json(method="service_add", name=name, item=service)

    def service_mod(self, name, service):
        return self._post_json(method="service_mod", name=name, item=service)

    def service_del(self, name):
        return self._post_json(method="service_del", name=name)

    def service_disable(self, name):
        return self._post_json(method="service_disable", name=name)

    def service_add_host(self, name, item):
        return self._post_json(method="service_add_host", name=name, item={"host": item})

    def service_remove_host(self, name, item):
        return self._post_json(method="service_remove_host", name=name, item={"host": item})


def get_service_dict(force=None, krbcanonicalname=None, skip_host_check=None):
    data = {}
    if force is not None:
        data["force"] = force
    if krbcanonicalname is not None:
        data["krbcanonicalname"] = krbcanonicalname
    if skip_host_check is not None:
        data["skip_host_check"] = skip_host_check
    return data


def get_service_diff(client, ipa_host, module_service):
    non_updateable_keys = ["force", "krbcanonicalname", "skip_host_check"]
    for key in non_updateable_keys:
        if key in module_service:
            del module_service[key]

    return client.get_diff(ipa_data=ipa_host, module_data=module_service)


def ensure(module, client):
    name = module.params["krbcanonicalname"]
    state = module.params["state"]
    hosts = module.params["hosts"]

    ipa_service = client.service_find(name=name)
    module_service = get_service_dict(force=module.params["force"], skip_host_check=module.params["skip_host_check"])
    changed = False
    if state in ["present", "enabled", "disabled"]:
        if not ipa_service:
            changed = True
            if not module.check_mode:
                client.service_add(name=name, service=module_service)
        else:
            diff = get_service_diff(client, ipa_service, module_service)
            if len(diff) > 0:
                changed = True
                if not module.check_mode:
                    data = {}
                    for key in diff:
                        data[key] = module_service.get(key)
                    client.service_mod(name=name, service=data)
        if hosts is not None:
            if "managedby_host" in ipa_service:
                for host in ipa_service["managedby_host"]:
                    if host not in hosts:
                        if not module.check_mode:
                            client.service_remove_host(name=name, item=host)
                        changed = True
                for host in hosts:
                    if host not in ipa_service["managedby_host"]:
                        if not module.check_mode:
                            client.service_add_host(name=name, item=host)
                        changed = True
            else:
                for host in hosts:
                    if not module.check_mode:
                        client.service_add_host(name=name, item=host)
                    changed = True

    else:
        if ipa_service:
            changed = True
            if not module.check_mode:
                client.service_del(name=name)

    return changed, client.service_find(name=name)


def main():
    argument_spec = ipa_argument_spec()
    argument_spec.update(
        krbcanonicalname=dict(type="str", required=True, aliases=["name"]),
        force=dict(type="bool"),
        skip_host_check=dict(type="bool", default=False),
        hosts=dict(type="list", elements="str"),
        state=dict(type="str", default="present", choices=["present", "absent"]),
    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    client = ServiceIPAClient(
        module=module,
        host=module.params["ipa_host"],
        port=module.params["ipa_port"],
        protocol=module.params["ipa_prot"],
    )

    try:
        client.login(username=module.params["ipa_user"], password=module.params["ipa_pass"])
        changed, host = ensure(module, client)
        module.exit_json(changed=changed, host=host)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


if __name__ == "__main__":
    main()



# Copyright (c) 2020, Jordan Borean <jborean93@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import enum
import re

from ansible.module_utils.common.text.converters import to_bytes

# An ASN.1 serialized as a string in the OpenSSL format:
#     [modifier,]type[:value]
#
# 'modifier':
#     The modifier can be 'IMPLICIT:<tag_number><tag_class>,' or 'EXPLICIT:<tag_number><tag_class>' where IMPLICIT
#     changes the tag of the universal value to encode and EXPLICIT prefixes its tag to the existing universal value.
#     The tag_number must be set while the tag_class can be 'U', 'A', 'P', or 'C" for 'Universal', 'Application',
#     'Private', or 'Context Specific' with C being the default.
#
# 'type':
#     The underlying ASN.1 type of the value specified. Currently only the following have been implemented:
#         UTF8: The value must be a UTF-8 encoded string.
#
# 'value':
#     The value to encode, the format of this value depends on the <type> specified.
ASN1_STRING_REGEX = re.compile(
    r"^((?P<tag_type>IMPLICIT|EXPLICIT):(?P<tag_number>\d+)(?P<tag_class>U|A|P|C)?,)?"
    r"(?P<value_type>[\w\d]+):(?P<value>.*)"
)


class TagClass(enum.Enum):
    UNIVERSAL = 0
    APPLICATION = 1
    CONTEXT_SPECIFIC = 2
    PRIVATE = 3


# Universal tag numbers that can be encoded.
class TagNumber(enum.Enum):
    UTF8_STRING = 12


def _pack_octet_integer(value: int) -> bytes:
    """Packs an integer value into 1 or multiple octets."""
    # NOTE: This is *NOT* the same as packing an ASN.1 INTEGER like value.
    octets = bytearray()

    # Continue to shift the number by 7 bits and pack into an octet until the
    # value is fully packed.
    while value:
        octet_value = value & 0b01111111

        # First round (last octet) must have the MSB set.
        if len(octets):
            octet_value |= 0b10000000

        octets.append(octet_value)
        value >>= 7

    # Reverse to ensure the higher order octets are first.
    octets.reverse()
    return bytes(octets)


def serialize_asn1_string_as_der(value: str) -> bytes:
    """Deserializes an ASN.1 string to a DER encoded byte string."""
    asn1_match = ASN1_STRING_REGEX.match(value)
    if not asn1_match:
        raise ValueError(
            "The ASN.1 serialized string must be in the format [modifier,]type[:value]"
        )

    tag_type = asn1_match.group("tag_type")
    tag_number = asn1_match.group("tag_number")
    tag_class = asn1_match.group("tag_class") or "C"
    value_type = asn1_match.group("value_type")
    asn1_value = asn1_match.group("value")

    if value_type != "UTF8":
        raise ValueError(
            f'The ASN.1 serialized string is not a known type "{value_type}", only UTF8 types are supported'
        )

    b_value = to_bytes(asn1_value, encoding="utf-8", errors="surrogate_or_strict")

    # We should only do a universal type tag if not IMPLICITLY tagged or the tag class is not universal.
    if not tag_type or (tag_type == "EXPLICIT" and tag_class != "U"):
        b_value = pack_asn1(
            tag_class=TagClass.UNIVERSAL,
            constructed=False,
            tag_number=TagNumber.UTF8_STRING,
            b_data=b_value,
        )

    if tag_type:
        tag_class_enum = {
            "U": TagClass.UNIVERSAL,
            "A": TagClass.APPLICATION,
            "P": TagClass.PRIVATE,
            "C": TagClass.CONTEXT_SPECIFIC,
        }[tag_class]

        # When adding support for more types this should be looked into further. For now it works with UTF8Strings.
        constructed = tag_type == "EXPLICIT" and tag_class_enum != TagClass.UNIVERSAL
        b_value = pack_asn1(
            tag_class=tag_class_enum,
            constructed=constructed,
            tag_number=int(tag_number),
            b_data=b_value,
        )

    return b_value


def pack_asn1(
    *,
    tag_class: TagClass,
    constructed: bool,
    tag_number: TagNumber | int,
    b_data: bytes,
) -> bytes:
    """Pack the value into an ASN.1 data structure.

    The structure for an ASN.1 element is

    | Identifier Octet(s) | Length Octet(s) | Data Octet(s) |
    """
    b_asn1_data = bytearray()

    # Bit 8 and 7 denotes the class.
    identifier_octets = tag_class.value << 6
    # Bit 6 denotes whether the value is primitive or constructed.
    identifier_octets |= (1 if constructed else 0) << 5

    # Bits 5-1 contain the tag number, if it cannot be encoded in these 5 bits
    # then they are set and another octet(s) is used to denote the tag number.
    if isinstance(tag_number, TagNumber):
        tag_number = tag_number.value
    if tag_number < 31:
        identifier_octets |= tag_number
        b_asn1_data.append(identifier_octets)
    else:
        identifier_octets |= 31
        b_asn1_data.append(identifier_octets)
        b_asn1_data.extend(_pack_octet_integer(tag_number))

    length = len(b_data)

    # If the length can be encoded in 7 bits only 1 octet is required.
    if length < 128:
        b_asn1_data.append(length)

    else:
        # Otherwise the length must be encoded across multiple octets
        length_octets = bytearray()
        while length:
            length_octets.append(length & 0b11111111)
            length >>= 8

        length_octets.reverse()  # Reverse to make the higher octets first.

        # The first length octet must have the MSB set alongside the number of
        # octets the length was encoded in.
        b_asn1_data.append(len(length_octets) | 0b10000000)
        b_asn1_data.extend(length_octets)

    return bytes(b_asn1_data) + b_data


__all__ = ("TagClass", "TagNumber", "pack_asn1", "serialize_asn1_string_as_der")



# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos l2_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""
import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import (
    utils,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.l2_interfaces.l2_interfaces import (
    L2_interfacesArgs,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.l2_interfaces import (
    L2_interfacesTemplate,
)


class L2_interfacesFacts(object):
    """The nxos l2_interfaces facts class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = L2_interfacesArgs.argument_spec

    def _get_interface_config(self, connection):
        return connection.get("show running-config | section ^interface")

    def _default_for_allowed_vlans(self, parsed_config):
        """Handle default for allowed vlans"""

        # Process allowed_vlans
        for interface in parsed_config:
            # if trunk and allowed vlan is empty then apply default
            # check if ...allowed vlan none command is not there
            # if vlan none command is there then don't apply default
            if interface.get("trunk"):
                if not interface.get("trunk", {}).get("allowed_vlans") and not interface.get(
                    "trunk",
                    {},
                ).get("allowed_vlans_none"):
                    interface["trunk"]["allowed_vlans"] = "1-4094"

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for L2_interfaces network resource

        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf

        :rtype: dictionary
        :returns: facts
        """
        facts = {}
        objs = []

        if not data:
            data = self._get_interface_config(connection)

        data = self._flatten_vlans(data)

        # parse native config using the L2_interfaces template
        l2_interfaces_parser = L2_interfacesTemplate(lines=data.splitlines(), module=self._module)
        objs = list(l2_interfaces_parser.parse().values())

        # process defaults for allowed vlan
        self._default_for_allowed_vlans(objs)

        ansible_facts["ansible_network_resources"].pop("l2_interfaces", None)

        params = utils.remove_empties(
            l2_interfaces_parser.validate_config(
                self.argument_spec,
                {"config": objs},
                redact=True,
            ),
        )

        facts["l2_interfaces"] = params.get("config", [])
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts

    def _flatten_vlans(self, data):
        """
        Flatten the vlan lines if it also contains lines that mention 'switchport trunk allowed vlan add'
        as we will merge all the entries in a list
        :param obj: data
        :returns: flattened vlan entries as switchport trunk allowed vlan
        """
        lines = data.split("\n")
        vlans = ""
        cur_indent = 0
        result = []
        regex_vlan_special_case = re.compile(
            r"\s+switchport\strunk\sallowed\svlan\s(none|all|except|remove)",
        )
        regex_vlan_add_line = re.compile(r"\s+switchport\strunk\sallowed\svlan\sadd")
        regex_vlan_set_line = re.compile(r"\s+switchport\strunk\sallowed\svlan")

        for line in lines:
            # If line starts with one of these entries, that is special one liner entry
            if regex_vlan_special_case.match(line):
                result.append(line)

            # If line starts with  allowed vlan add
            elif regex_vlan_add_line.match(line):
                if vlans:
                    vlans += "," + line.rsplit("add", maxsplit=1)[-1].strip()
                else:
                    vlans = line.rsplit("add", maxsplit=1)[-1].strip()
                cur_indent = len(line) - len(line.lstrip())

            # If line starts only with allowed vlan
            elif regex_vlan_set_line.match(line):
                vlans = line.rsplit("vlan", maxsplit=1)[-1].strip()
                cur_indent = len(line) - len(line.lstrip())

            else:
                if vlans:
                    result.append(f"{' ' * cur_indent}switchport trunk allowed vlan {vlans}")
                    vlans = ""
                result.append(line)

        if vlans:
            result.append(f"{' ' * cur_indent}switchport trunk allowed vlan {vlans}")

        return "\n".join(result)



#!/usr/bin/python
# Copyright (c) Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: get_certificate
author: "John Westcott IV (@john-westcott-iv)"
short_description: Get a certificate from a host:port
description:
  - Makes a secure connection and returns information about the presented certificate.
  - The module uses the cryptography Python library.
extends_documentation_fragment:
  - community.crypto._attributes
  - community.crypto._attributes.idempotent_not_modify_state
  - community.crypto._cryptography_dep.minimum
attributes:
  check_mode:
    support: none
    details:
      - This action does not modify state.
  diff_mode:
    support: N/A
    details:
      - This action does not modify state.
options:
  host:
    description:
      - The host to get the cert for (IP is fine).
    type: str
    required: true
  ca_cert:
    description:
      - A PEM file containing one or more root certificates; if present, the cert will be validated against these root certs.
      - Note that this only validates the certificate is signed by the chain; not that the cert is valid for the host presenting
        it.
    type: path
  port:
    description:
      - The port to connect to.
    type: int
    required: true
  server_name:
    description:
      - Server name used for SNI (L(Server Name Indication,https://en.wikipedia.org/wiki/Server_Name_Indication)) when hostname
        is an IP or is different from server name.
    type: str
    version_added: 1.4.0
  proxy_host:
    description:
      - Proxy host used when get a certificate.
    type: str
  proxy_port:
    description:
      - Proxy port used when get a certificate.
    type: int
    default: 8080
  starttls:
    description:
      - Requests a secure connection for protocols which require clients to initiate encryption.
      - Only available for V(mysql) currently.
    type: str
    choices:
      - mysql
    version_added: 1.9.0
  timeout:
    description:
      - The timeout in seconds.
    type: int
    default: 10
  select_crypto_backend:
    description:
      - Determines which crypto backend to use.
      - The default choice is V(auto), which tries to use C(cryptography) if available.
      - If set to V(cryptography), will try to use the L(cryptography,https://cryptography.io/) library.
      - Note that with community.crypto 3.0.0, all values behave the same.
        This option will be deprecated in a later version.
        We recommend to not set it explicitly.
    type: str
    default: auto
    choices: [auto, cryptography]
  ciphers:
    description:
      - SSL/TLS Ciphers to use for the request.
      - When a list is provided, all ciphers are joined in order with V(:).
      - See the L(OpenSSL Cipher List Format,https://www.openssl.org/docs/manmaster/man1/openssl-ciphers.html#CIPHER-LIST-FORMAT)
        for more details.
      - The available ciphers is dependent on the Python and OpenSSL/LibreSSL versions.
    type: list
    elements: str
    version_added: 2.11.0
  asn1_base64:
    description:
      - Whether to encode the ASN.1 values in the RV(extensions) return value with Base64 or not.
      - The documentation claimed for a long time that the values are Base64 encoded, but they never were. For compatibility
        this option is set to V(false).
      - The default value was changed from V(false) to V(true) incommunity.crypto 3.0.0.
    type: bool
    default: true
    version_added: 2.12.0
  tls_ctx_options:
    description:
      - TLS context options (TLS/SSL OP flags) to use for the request.
      - See the L(List of SSL OP Flags,https://wiki.openssl.org/index.php/List_of_SSL_OP_Flags) for more details.
      - The available TLS context options is dependent on the Python and OpenSSL/LibreSSL versions.
    type: list
    elements: raw
    version_added: 2.21.0
  get_certificate_chain:
    description:
      - If set to V(true), will obtain the certificate chain next to the certificate itself.
      - The chain as returned by the server can be found in RV(unverified_chain), and the chain that passed validation in
        RV(verified_chain).
      - B(Note) that this needs B(Python 3.10 or newer). Also note that only Python 3.13 or newer officially supports this.
        The module uses internal APIs of Python 3.10, 3.11, and 3.12 to achieve the same. It can be that future versions of
        Python 3.10, 3.11, or 3.12 break this.
    type: bool
    default: false
    version_added: 2.21.0

notes:
  - When using ca_cert on OS X it has been reported that in some conditions the validate will always succeed.
requirements:
  - "Python >= 3.10 when O(get_certificate_chain=true)"

seealso:
  - plugin: community.crypto.to_serial
    plugin_type: filter
"""

RETURN = r"""
cert:
  description: The certificate retrieved from the port.
  returned: success
  type: str
expired:
  description: Boolean indicating if the cert is expired.
  returned: success
  type: bool
extensions:
  description: Extensions applied to the cert.
  returned: success
  type: list
  elements: dict
  contains:
    critical:
      returned: success
      type: bool
      description: Whether the extension is critical.
    asn1_data:
      returned: success
      type: str
      description:
        - The ASN.1 content of the extension.
        - If O(asn1_base64=true) this will be Base64 encoded, otherwise the raw binary value will be returned.
        - Please note that the raw binary value might not survive JSON serialization to the Ansible controller, and also might
          cause failures when displaying it. See U(https://github.com/ansible/ansible/issues/80258) for more information.
        - B(Note) that depending on the C(cryptography) version used, it is not possible to extract the ASN.1 content of the
          extension, but only to provide the re-encoded content of the extension in case it was parsed by C(cryptography).
          This should usually result in exactly the same value, except if the original extension value was malformed.
    name:
      returned: success
      type: str
      description: The extension's name.
issuer:
  description: Information about the issuer of the cert.
  returned: success
  type: dict
not_after:
  description: Expiration date of the cert.
  returned: success
  type: str
not_before:
  description: Issue date of the cert.
  returned: success
  type: str
serial_number:
  description:
    - The serial number of the cert.
    - This return value is an B(integer). If you need the serial numbers as a colon-separated hex string, such as C(11:22:33),
      you need to convert it to that form with P(community.crypto.to_serial#filter).
  returned: success
  type: int
signature_algorithm:
  description: The algorithm used to sign the cert.
  returned: success
  type: str
subject:
  description: Information about the subject of the cert (C(OU), C(CN), and so on).
  returned: success
  type: dict
version:
  description: The version number of the certificate.
  returned: success
  type: str
verified_chain:
  description:
    - The verified certificate chain retrieved from the port.
    - The first entry is always RV(cert).
    - The last certificate the root certificate the chain is traced to. If O(ca_cert) is provided this certificate is part
      of that store; otherwise it is part of the store used by default by Python.
    - Note that RV(unverified_chain) generally does not contain the root certificate, and might contain other certificates
      that are not part of the validated chain.
  returned: success and O(get_certificate_chain=true)
  type: list
  elements: str
  version_added: 2.21.0
unverified_chain:
  description:
    - The certificate chain retrieved from the port.
    - The first entry is always RV(cert).
  returned: success and O(get_certificate_chain=true)
  type: list
  elements: str
  version_added: 2.21.0
"""

EXAMPLES = r"""
---
- name: Get the cert from an RDP port
  community.crypto.get_certificate:
    host: "1.2.3.4"
    port: 3389
  delegate_to: localhost
  run_once: true
  register: cert

- name: Get a cert from an https port
  community.crypto.get_certificate:
    host: "www.google.com"
    port: 443
  delegate_to: localhost
  run_once: true
  register: cert

- name: How many days until cert expires
  ansible.builtin.debug:
    msg: "cert expires in: {{ expire_days }} days."
  vars:
    expire_days: >-
      {{ (
        (cert.not_after | ansible.builtin.to_datetime('%Y%m%d%H%M%SZ')) -
        (ansible_date_time.iso8601 | ansible.builtin.to_datetime('%Y-%m-%dT%H:%M:%SZ'))
      ).days }}

- name: Allow legacy insecure renegotiation to get a cert from a legacy device
  community.crypto.get_certificate:
    host: "legacy-device.domain.com"
    port: 443
    ciphers:
      - HIGH
    tls_ctx_options:
      - OP_ALL
      - OP_NO_SSLv3
      - OP_CIPHER_SERVER_PREFERENCE
      - OP_ENABLE_MIDDLEBOX_COMPAT
      - OP_NO_COMPRESSION
      - 4 # OP_LEGACY_SERVER_CONNECT
  delegate_to: localhost
  run_once: true
  register: legacy_cert
"""

import atexit
import base64
import ssl
import sys
import typing as t
from os.path import isfile
from socket import create_connection, setdefaulttimeout, socket
from ssl import (
    CERT_NONE,
    CERT_REQUIRED,
    DER_cert_to_PEM_cert,
    create_default_context,
)

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_bytes, to_text

from ansible_collections.community.crypto.plugins.module_utils._crypto.cryptography_support import (
    CRYPTOGRAPHY_TIMEZONE,
    cryptography_get_extensions_from_cert,
    cryptography_oid_to_name,
    get_not_valid_after,
    get_not_valid_before,
)
from ansible_collections.community.crypto.plugins.module_utils._cryptography_dep import (
    COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION,
    assert_required_cryptography_version,
)
from ansible_collections.community.crypto.plugins.module_utils._time import (
    get_now_datetime,
)

MINIMAL_CRYPTOGRAPHY_VERSION = COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION

try:
    import cryptography
    import cryptography.exceptions
    import cryptography.x509
except ImportError:
    pass


def send_starttls_packet(sock: socket, server_type: t.Literal["mysql"]) -> None:
    if server_type == "mysql":
        ssl_request_packet = (
            b"\x20\x00\x00\x01\x85\xae\x7f\x00"
            + b"\x00\x00\x00\x01\x21\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00"
        )

        sock.recv(
            8192
        )  # discard initial handshake from server for this naive implementation
        sock.send(ssl_request_packet)


def main() -> t.NoReturn:
    module = AnsibleModule(
        argument_spec={
            "ca_cert": {"type": "path"},
            "host": {"type": "str", "required": True},
            "port": {"type": "int", "required": True},
            "proxy_host": {"type": "str"},
            "proxy_port": {"type": "int", "default": 8080},
            "server_name": {"type": "str"},
            "timeout": {"type": "int", "default": 10},
            "select_crypto_backend": {
                "type": "str",
                "default": "auto",
                "choices": ["auto", "cryptography"],
            },
            "starttls": {"type": "str", "choices": ["mysql"]},
            "ciphers": {"type": "list", "elements": "str"},
            "asn1_base64": {"type": "bool", "default": True},
            "tls_ctx_options": {"type": "list", "elements": "raw"},
            "get_certificate_chain": {"type": "bool", "default": False},
        },
    )

    ca_cert: str | None = module.params.get("ca_cert")
    host: str = module.params.get("host")
    port: int = module.params.get("port")
    proxy_host: str | None = module.params.get("proxy_host")
    proxy_port: int | None = module.params.get("proxy_port")
    timeout: int = module.params.get("timeout")
    server_name: str | None = module.params.get("server_name")
    start_tls_server_type: t.Literal["mysql"] | None = module.params.get("starttls")
    ciphers: list[str] | None = module.params.get("ciphers")
    asn1_base64: bool = module.params["asn1_base64"]
    tls_ctx_options: list[str | bytes | int] | None = module.params["tls_ctx_options"]
    get_certificate_chain: bool = module.params["get_certificate_chain"]

    if get_certificate_chain and sys.version_info < (3, 10):
        module.fail_json(
            msg="get_certificate_chain=true can only be used with Python 3.10 (Python 3.13+ officially supports this). "
            f"The Python version used to run the get_certificate module is {sys.version}"
        )

    assert_required_cryptography_version(
        module, minimum_cryptography_version=MINIMAL_CRYPTOGRAPHY_VERSION
    )

    result: dict[str, t.Any] = {
        "changed": False,
    }

    if timeout:
        setdefaulttimeout(timeout)

    if ca_cert and not isfile(ca_cert):
        module.fail_json(msg="ca_cert file does not exist")

    verified_chain = None
    unverified_chain = None

    try:
        if proxy_host:
            connect = f"CONNECT {host}:{port} HTTP/1.0\r\n\r\n"
            sock = socket()
            atexit.register(sock.close)
            sock.connect((proxy_host, proxy_port))
            sock.send(connect.encode())
            sock.recv(8192)
        else:
            sock = create_connection((host, port))
            atexit.register(sock.close)

        if ca_cert:
            ctx = create_default_context(cafile=ca_cert)
            ctx.check_hostname = False
            ctx.verify_mode = CERT_REQUIRED
        else:
            ctx = create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = CERT_NONE

        if start_tls_server_type is not None:
            send_starttls_packet(sock, start_tls_server_type)

        if ciphers is not None:
            ciphers_joined = ":".join(ciphers)
            ctx.set_ciphers(ciphers_joined)

        if tls_ctx_options is not None:
            # Clear default ctx options
            ctx.options = 0  # type: ignore

            # For each item in the tls_ctx_options list
            for tls_ctx_option in tls_ctx_options:
                # If the item is a string_type
                if isinstance(tls_ctx_option, (str, bytes)):
                    # Convert tls_ctx_option to a native string
                    tls_ctx_option_str = to_text(tls_ctx_option)
                    # Get the tls_ctx_option_str attribute from ssl
                    tls_ctx_option_attr = getattr(ssl, tls_ctx_option_str, None)
                    # If tls_ctx_option_attr is an integer
                    if isinstance(tls_ctx_option_attr, int):
                        # Set tls_ctx_option_int to the attribute value
                        tls_ctx_option_int = tls_ctx_option_attr
                    # If tls_ctx_option_attr is not an integer
                    else:
                        module.fail_json(
                            msg=f"Failed to determine the numeric value for {tls_ctx_option_str}"
                        )
                # If the item is an integer
                elif isinstance(tls_ctx_option, int):
                    # Set tls_ctx_option_int to the item value
                    tls_ctx_option_int = tls_ctx_option
                # If the item is not a string nor integer
                else:
                    module.fail_json(
                        msg=f"tls_ctx_options must be a string or integer, got {tls_ctx_option!r}"
                    )
                    tls_ctx_option_int = (  # type: ignore[unreachable]
                        0  # make pylint happy; this code is actually unreachable
                    )

                try:
                    # Add the int value of the item to ctx options
                    # (pylint does not yet notice that module.fail_json cannot return)
                    ctx.options |= tls_ctx_option_int  # pylint: disable=possibly-used-before-assignment
                except Exception:
                    module.fail_json(
                        msg=f"Failed to add {tls_ctx_option_str or tls_ctx_option_int} to CTX options"
                    )

        tls_sock = ctx.wrap_socket(sock, server_hostname=server_name or host)
        cert_der = tls_sock.getpeercert(True)
        if cert_der is None:
            raise Exception("Unexpected error: no peer certificate has been returned")
        cert: str = DER_cert_to_PEM_cert(cert_der)

        if get_certificate_chain:
            if sys.version_info < (3, 13):
                # The official way to access this has been added in https://github.com/python/cpython/pull/109113/files.
                # We are basically doing the same for older Python versions. The internal API needed for this was added
                # in https://github.com/python/cpython/commit/666991fc598bc312d72aff0078ecb553f0a968f1, which was first
                # released in Python 3.10.0.
                def _convert_chain(chain):
                    if not chain:
                        return []
                    return [
                        c.public_bytes(
                            ssl._ssl.ENCODING_DER  # pylint: disable=protected-access
                        )
                        for c in chain
                    ]

                ssl_obj = (
                    tls_sock._sslobj  # pylint: disable=protected-access
                )  # This is of type ssl._ssl._SSLSocket
                verified_der_chain = _convert_chain(ssl_obj.get_verified_chain())
                unverified_der_chain = _convert_chain(ssl_obj.get_unverified_chain())
            else:
                # This works with Python 3.13+

                # Unfortunately due to a bug (https://github.com/python/cpython/issues/118658) some early pre-releases of
                # Python 3.13 do not return lists of byte strings, but lists of _ssl.Certificate objects. This is going to
                # be fixed by https://github.com/python/cpython/pull/118669. For now we convert the certificates ourselves
                # if they are not byte strings to work around this.
                def _convert_chain(chain: list[bytes]) -> list[bytes]:
                    return [
                        (
                            c
                            if isinstance(c, bytes)
                            else c.public_bytes(
                                ssl._ssl.ENCODING_DER  # pylint: disable=protected-access
                            )
                        )
                        for c in chain
                    ]

                verified_der_chain = _convert_chain(tls_sock.get_verified_chain())
                unverified_der_chain = _convert_chain(tls_sock.get_unverified_chain())

            verified_chain = [DER_cert_to_PEM_cert(c) for c in verified_der_chain]
            unverified_chain = [DER_cert_to_PEM_cert(c) for c in unverified_der_chain]

    except Exception as e:
        if proxy_host:
            module.fail_json(
                msg=f"Failed to get cert via proxy {proxy_host}:{proxy_port} from {host}:{port}, error: {e}"
            )
        else:
            module.fail_json(msg=f"Failed to get cert from {host}:{port}, error: {e}")

    result["cert"] = cert

    x509 = cryptography.x509.load_pem_x509_certificate(to_bytes(cert))
    result["subject"] = {}
    for attribute in x509.subject:
        result["subject"][cryptography_oid_to_name(attribute.oid, short=True)] = (
            attribute.value
        )

    result["expired"] = get_not_valid_after(x509) < get_now_datetime(
        with_timezone=CRYPTOGRAPHY_TIMEZONE
    )

    result["extensions"] = []
    for dotted_number, entry in cryptography_get_extensions_from_cert(x509).items():
        oid = cryptography.x509.oid.ObjectIdentifier(dotted_number)
        ext: dict[str, t.Any] = {
            "critical": entry["critical"],
            "asn1_data": entry["value"],
            "name": cryptography_oid_to_name(oid, short=True),
        }
        if not asn1_base64:
            ext["asn1_data"] = base64.b64decode(entry["value"])  # type: ignore
        result["extensions"].append(ext)

    result["issuer"] = {}
    for attribute in x509.issuer:
        result["issuer"][cryptography_oid_to_name(attribute.oid, short=True)] = (
            attribute.value
        )

    result["not_after"] = get_not_valid_after(x509).strftime("%Y%m%d%H%M%SZ")
    result["not_before"] = get_not_valid_before(x509).strftime("%Y%m%d%H%M%SZ")

    result["serial_number"] = x509.serial_number
    result["signature_algorithm"] = cryptography_oid_to_name(
        x509.signature_algorithm_oid
    )

    # We need the -1 offset to get the same values as pyOpenSSL
    if x509.version == cryptography.x509.Version.v1:
        result["version"] = 1 - 1
    elif x509.version == cryptography.x509.Version.v3:
        result["version"] = 3 - 1
    else:
        result["version"] = "unknown"  # type: ignore[unreachable]

    if verified_chain is not None:
        result["verified_chain"] = verified_chain
    if unverified_chain is not None:
        result["unverified_chain"] = unverified_chain

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2016, Josh Bradley <jbradley(at)digitalocean.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: chef_databag
short_description: Fetches data from a Chef Databag
description:
  - 'This is a lookup plugin to provide access to chef data bags using the pychef package. It interfaces with the chef server
    API using the same methods to find a knife or chef-client config file to load parameters from, starting from either the
    given base path or the current working directory. The lookup order mirrors the one from Chef, all folders in the base
    path are walked back looking for the following configuration file in order: C(.chef/knife.rb), C(~/.chef/knife.rb), C(/etc/chef/client.rb).'
requirements:
  - "pychef (L(Python library, https://pychef.readthedocs.io), C(pip install pychef))"
options:
  name:
    description:
      - Name of the databag.
    type: string
    required: true
  item:
    description:
      - Item to fetch.
    type: string
    required: true
"""

EXAMPLES = r"""
- ansible.builtin.debug:
    msg: "{{ lookup('community.general.chef_databag', 'name=data_bag_name item=data_bag_item') }}"
"""

RETURN = r"""
_raw:
  description:
    - The value from the databag.
  type: list
  elements: dict
"""

from ansible.errors import AnsibleError
from ansible.parsing.splitter import parse_kv
from ansible.plugins.lookup import LookupBase

try:
    import chef

    HAS_CHEF = True
except ImportError:
    HAS_CHEF = False


class LookupModule(LookupBase):
    """
    Chef data bag lookup module
    """

    def __init__(self, loader=None, templar=None, **kwargs):
        super().__init__(loader, templar, **kwargs)

        # setup vars for data bag name and data bag item
        self.name = None
        self.item = None

    def parse_kv_args(self, args):
        """
        parse key-value style arguments
        """

        for arg in ["name", "item"]:
            try:
                arg_raw = args.pop(arg, None)
                if arg_raw is None:
                    continue
                parsed = str(arg_raw)
                setattr(self, arg, parsed)
            except ValueError as e:
                raise AnsibleError(f"can't parse arg {arg}={arg_raw} as string") from e
        if args:
            raise AnsibleError(f"unrecognized arguments to with_sequence: {list(args.keys())!r}")

    def run(self, terms, variables=None, **kwargs):
        # Ensure pychef has been loaded
        if not HAS_CHEF:
            raise AnsibleError("PyChef needed for lookup plugin, try `pip install pychef`")

        # TODO: use new-style option parsing

        for term in terms:
            self.parse_kv_args(parse_kv(term))

        api_object = chef.autoconfigure()

        if not isinstance(api_object, chef.api.ChefAPI):
            raise AnsibleError("Unable to connect to Chef Server API.")

        data_bag_object = chef.DataBag(self.name)

        data_bag_item = data_bag_object[self.item]

        return [dict(data_bag_item)]



#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_userprofile

short_description: Allows managing Keycloak User Profiles

description:
  - This module allows you to create, update, or delete Keycloak User Profiles using the Keycloak API. You can also customize
    the "Unmanaged Attributes" with it.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/24.0.5/rest-api/index.html). For compatibility reasons, the module also accepts
    the camelCase versions of the options.
version_added: "9.4.0"

attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the User Profile provider.
      - On V(present), the User Profile provider is created if it does not yet exist, or updated with the parameters you provide.
      - On V(absent), the User Profile provider is removed if it exists.
    default: 'present'
    type: str
    choices:
      - present
      - absent

  parent_id:
    description:
      - The parent ID of the realm key. In practice the ID (name) of the realm.
    aliases:
      - parentId
      - realm
    type: str
    required: true

  provider_id:
    description:
      - The name of the provider ID for the key (supported value is V(declarative-user-profile)).
    aliases:
      - providerId
    choices: ['declarative-user-profile']
    default: 'declarative-user-profile'
    type: str

  provider_type:
    description:
      - Component type for User Profile (only supported value is V(org.keycloak.userprofile.UserProfileProvider)).
    aliases:
      - providerType
    choices: ['org.keycloak.userprofile.UserProfileProvider']
    default: org.keycloak.userprofile.UserProfileProvider
    type: str

  config:
    description:
      - The configuration of the User Profile Provider.
    type: dict
    suboptions:
      kc_user_profile_config:
        description:
          - Define a declarative User Profile. See EXAMPLES for more context.
        aliases:
          - kcUserProfileConfig
        type: list
        elements: dict
        suboptions:
          attributes:
            description:
              - A list of attributes to be included in the User Profile.
            type: list
            elements: dict
            suboptions:
              name:
                description:
                  - The name of the attribute.
                type: str
                required: true

              display_name:
                description:
                  - The display name of the attribute.
                aliases:
                  - displayName
                type: str
                required: true

              validations:
                description:
                  - The validations to be applied to the attribute.
                type: dict
                suboptions:
                  length:
                    description:
                      - The length validation for the attribute.
                    type: dict
                    suboptions:
                      min:
                        description:
                          - The minimum length of the attribute.
                        type: int
                      max:
                        description:
                          - The maximum length of the attribute.
                        type: int
                        required: true

                  email:
                    description:
                      - The email validation for the attribute.
                    type: dict

                  username_prohibited_characters:
                    description:
                      - The prohibited characters validation for the username attribute.
                    type: dict
                    aliases:
                      - usernameProhibitedCharacters

                  up_username_not_idn_homograph:
                    description:
                      - The validation to prevent IDN homograph attacks in usernames.
                    type: dict
                    aliases:
                      - upUsernameNotIdnHomograph

                  person_name_prohibited_characters:
                    description:
                      - The prohibited characters validation for person name attributes.
                    type: dict
                    aliases:
                      - personNameProhibitedCharacters

                  uri:
                    description:
                      - The URI validation for the attribute.
                    type: dict

                  pattern:
                    description:
                      - The pattern validation for the attribute using regular expressions.
                    type: dict

                  options:
                    description:
                      - Validation to ensure the attribute matches one of the provided options.
                    type: dict

                  integer:
                    description:
                      - The integer validation for the attribute.
                    type: dict
                    version_added: 12.2.0

                  double:
                    description:
                      - The double validation for the attribute.
                    type: dict
                    version_added: 12.2.0

                  iso_date:
                    description:
                      - The iso-date validation for the attribute.
                    type: dict
                    aliases:
                      - isoDate
                    version_added: 12.2.0

                  local_date:
                    description:
                      - The local-date validation for the attribute.
                    type: dict
                    aliases:
                      - localDate
                    version_added: 12.2.0

                  multivalued:
                    description:
                      - The multivalued validation for the attribute.
                    type: dict
                    suboptions:
                      min:
                        description:
                          - The minimum amount of values of the attribute.
                        type: int
                      max:
                        description:
                          - The maximum amount of values of the attribute.
                        type: int
                        required: true
                    version_added: 12.2.0

              annotations:
                description:
                  - Annotations for the attribute.
                type: dict

              group:
                description:
                  - Specifies the User Profile group where this attribute is added.
                type: str

              permissions:
                description:
                  - The permissions for viewing and editing the attribute.
                type: dict
                suboptions:
                  view:
                    description:
                      - The roles that can view the attribute.
                      - Supported values are V(admin) and V(user).
                    type: list
                    elements: str
                    default:
                      - admin
                      - user

                  edit:
                    description:
                      - The roles that can edit the attribute.
                      - Supported values are V(admin) and V(user).
                    type: list
                    elements: str
                    default:
                      - admin
                      - user

              multivalued:
                description:
                  - Whether the attribute can have multiple values.
                type: bool
                default: false

              required:
                description:
                  - The roles that require this attribute.
                type: dict
                suboptions:
                  roles:
                    description:
                      - The roles for which this attribute is required.
                      - Supported values are V(admin) and V(user).
                    type: list
                    elements: str
                    default:
                      - user

              selector:
                description:
                  - Selector when the attribute should be added.
                type: dict
                version_added: 12.2.0
                suboptions:
                  scopes:
                    description:
                      - Scopes to which the attribute should be added.
                    type: list
                    elements: str

          groups:
            description:
              - A list of attribute groups to be included in the User Profile.
            type: list
            elements: dict
            suboptions:
              name:
                description:
                  - The name of the group.
                type: str
                required: true

              display_header:
                description:
                  - The display header for the group.
                aliases:
                  - displayHeader
                type: str
                required: true

              display_description:
                description:
                  - The display description for the group.
                aliases:
                  - displayDescription
                type: str

              annotations:
                description:
                  - The annotations included in the group.
                type: dict

          unmanaged_attribute_policy:
            description:
              - Policy for unmanaged attributes.
            aliases:
              - unmanagedAttributePolicy
            type: str
            choices:
              - ENABLED
              - ADMIN_EDIT
              - ADMIN_VIEW

notes:
  - Currently, only a single V(declarative-user-profile) entry is supported for O(provider_id) (design of the Keyckoak API).
    However, there can be multiple O(config.kc_user_profile_config[].attributes[]) entries.
extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Eike Waldt (@yeoldegrove)
"""

EXAMPLES = r"""
- name: Create a Declarative User Profile with default settings
  community.general.keycloak_userprofile:
    state: present
    parent_id: master
    config:
      kc_user_profile_config:
        - attributes:
            - name: username
              displayName: ${username}
              validations:
                length:
                  min: 3
                  max: 255
                username_prohibited_characters: {}
                up_username_not_idn_homograph: {}
              annotations: {}
              permissions:
                view:
                  - admin
                  - user
                edit: []
              multivalued: false
            - name: email
              displayName: ${email}
              validations:
                email: {}
                length:
                  max: 255
              annotations: {}
              required:
                roles:
                  - user
              permissions:
                view:
                  - admin
                  - user
                edit: []
              multivalued: false
            - name: firstName
              displayName: ${firstName}
              validations:
                length:
                  max: 255
                person_name_prohibited_characters: {}
              annotations: {}
              required:
                roles:
                  - user
              permissions:
                view:
                  - admin
                  - user
                edit: []
              multivalued: false
            - name: lastName
              displayName: ${lastName}
              validations:
                length:
                  max: 255
                person_name_prohibited_characters: {}
              annotations: {}
              required:
                roles:
                  - user
              permissions:
                view:
                  - admin
                  - user
                edit: []
              multivalued: false
            - name: testAttribute
              displayName: ${testAttribute}
              validations:
                integer:
                  min: 0
                  max: 255
              annotations: {}
              required:
                roles:
                  - user
              permissions:
                view:
                  - admin
                  - user
                edit: []
              multivalued: false
          groups:
            - name: user-metadata
              displayHeader: User metadata
              displayDescription: Attributes, which refer to user metadata
              annotations: {}

- name: Delete a Keycloak User Profile Provider
  keycloak_userprofile:
    state: absent
    parent_id: master

# Unmanaged attributes are user attributes not explicitly defined in the User Profile
# configuration. By default, unmanaged attributes are "Disabled" and are not
# available from any context such as registration, account, and the
# administration console. By setting "Enabled", unmanaged attributes are fully
# recognized by the server and accessible through all contexts, useful if you are
# starting migrating an existing realm to the declarative User Profile
# and you don't have yet all user attributes defined in the User Profile configuration.
- name: Enable Unmanaged Attributes
  community.general.keycloak_userprofile:
    state: present
    parent_id: master
    config:
      kc_user_profile_config:
        - unmanagedAttributePolicy: ENABLED

# By setting "Only administrators can write", unmanaged attributes can be managed
# only through the administration console and API, useful if you have already
# defined any custom attribute that can be managed by users but you are unsure
# about adding other attributes that should only be managed by administrators.
- name: Enable ADMIN_EDIT on Unmanaged Attributes
  community.general.keycloak_userprofile:
    state: present
    parent_id: master
    config:
      kc_user_profile_config:
        - unmanagedAttributePolicy: ADMIN_EDIT

# By setting `Only administrators can view`, unmanaged attributes are read-only
# and only available through the administration console and API.
- name: Enable ADMIN_VIEW on Unmanaged Attributes
  community.general.keycloak_userprofile:
    state: present
    parent_id: master
    config:
      kc_user_profile_config:
        - unmanagedAttributePolicy: ADMIN_VIEW
"""

RETURN = r"""
msg:
  description: The output message generated by the module.
  returned: always
  type: str
  sample: UserProfileProvider created successfully
data:
  description: The data returned by the Keycloak API.
  returned: when state is present
  type: dict
"""

import json
from copy import deepcopy
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    camel,
    get_token,
    keycloak_argument_spec,
)


def remove_null_values(data):
    if isinstance(data, dict):
        # Recursively remove null values from dictionaries
        return {k: remove_null_values(v) for k, v in data.items() if v is not None}
    elif isinstance(data, list):
        # Recursively remove null values from lists
        return [remove_null_values(item) for item in data if item is not None]
    else:
        # Return the data if it is neither a dictionary nor a list
        return data


def camel_recursive(data):
    if isinstance(data, dict):
        # Convert keys to camelCase and apply recursively
        return {camel(k): camel_recursive(v) for k, v in data.items()}
    elif isinstance(data, list):
        # Apply camelCase conversion to each item in the list
        return [camel_recursive(item) for item in data]
    else:
        # Return the data as-is if it is not a dict or list
        return data


def main():
    argument_spec = keycloak_argument_spec()

    meta_args = dict(
        state=dict(type="str", choices=["present", "absent"], default="present"),
        parent_id=dict(type="str", aliases=["parentId", "realm"], required=True),
        provider_id=dict(
            type="str", aliases=["providerId"], default="declarative-user-profile", choices=["declarative-user-profile"]
        ),
        provider_type=dict(
            type="str",
            aliases=["providerType"],
            default="org.keycloak.userprofile.UserProfileProvider",
            choices=["org.keycloak.userprofile.UserProfileProvider"],
        ),
        config=dict(
            type="dict",
            options={
                "kc_user_profile_config": dict(
                    type="list",
                    aliases=["kcUserProfileConfig"],
                    elements="dict",
                    options={
                        "attributes": dict(
                            type="list",
                            elements="dict",
                            options={
                                "name": dict(type="str", required=True),
                                "display_name": dict(type="str", aliases=["displayName"], required=True),
                                "validations": dict(
                                    type="dict",
                                    options={
                                        "length": dict(
                                            type="dict",
                                            options={"min": dict(type="int"), "max": dict(type="int", required=True)},
                                        ),
                                        "email": dict(type="dict"),
                                        "username_prohibited_characters": dict(
                                            type="dict", aliases=["usernameProhibitedCharacters"]
                                        ),
                                        "up_username_not_idn_homograph": dict(
                                            type="dict", aliases=["upUsernameNotIdnHomograph"]
                                        ),
                                        "person_name_prohibited_characters": dict(
                                            type="dict", aliases=["personNameProhibitedCharacters"]
                                        ),
                                        "uri": dict(type="dict"),
                                        "pattern": dict(type="dict"),
                                        "options": dict(type="dict"),
                                        "integer": dict(type="dict"),
                                        "double": dict(type="dict"),
                                        "iso_date": dict(type="dict", aliases=["isoDate"]),
                                        "local_date": dict(type="dict", aliases=["localDate"]),
                                        "multivalued": dict(
                                            type="dict",
                                            options={
                                                "min": dict(type="int", required=False),
                                                "max": dict(type="int", required=True),
                                            },
                                        ),
                                    },
                                ),
                                "annotations": dict(type="dict"),
                                "group": dict(type="str"),
                                "permissions": dict(
                                    type="dict",
                                    options={
                                        "view": dict(type="list", elements="str", default=["admin", "user"]),
                                        "edit": dict(type="list", elements="str", default=["admin", "user"]),
                                    },
                                ),
                                "multivalued": dict(type="bool", default=False),
                                "required": dict(
                                    type="dict", options={"roles": dict(type="list", elements="str", default=["user"])}
                                ),
                                "selector": dict(type="dict", options={"scopes": dict(type="list", elements="str")}),
                            },
                        ),
                        "groups": dict(
                            type="list",
                            elements="dict",
                            options={
                                "name": dict(type="str", required=True),
                                "display_header": dict(type="str", aliases=["displayHeader"], required=True),
                                "display_description": dict(type="str", aliases=["displayDescription"]),
                                "annotations": dict(type="dict"),
                            },
                        ),
                        "unmanaged_attribute_policy": dict(
                            type="str",
                            aliases=["unmanagedAttributePolicy"],
                            choices=["ENABLED", "ADMIN_EDIT", "ADMIN_VIEW"],
                        ),
                    },
                )
            },
        ),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"]]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    # Initialize the result object. Only "changed" seems to have special
    # meaning for Ansible.
    result = dict(changed=False, msg="", end_state={}, diff=dict(before={}, after={}))

    # This will include the current state of the realm userprofile if it is already
    # present. This is only used for diff-mode.
    before_realm_userprofile = {}
    before_realm_userprofile["config"] = {}

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    params_to_ignore = list(keycloak_argument_spec().keys()) + ["state"]

    # Filter and map the parameters names that apply to the role
    component_params = [x for x in module.params if x not in params_to_ignore and module.params.get(x) is not None]

    # Build a proposed changeset from parameters given to this module
    changeset = {}

    # Build the changeset with proper JSON serialization for kc_user_profile_config
    config = module.params.get("config")
    changeset["config"] = {}

    # Generate a JSON payload for Keycloak Admin API from the module
    # parameters.  Parameters that do not belong to the JSON payload (e.g.
    # "state" or "auth_keycloal_url") have been filtered away earlier (see
    # above).
    #
    # This loop converts Ansible module parameters (snake-case) into
    # Keycloak-compatible format (camel-case). For example proider_id
    # becomes providerId. It also handles some special cases, e.g. aliases.
    for component_param in component_params:
        # realm/parent_id parameter
        if component_param == "realm" or component_param == "parent_id":
            changeset["parent_id"] = module.params.get(component_param)
            changeset.pop(component_param, None)
        # complex parameters in config suboptions
        elif component_param == "config":
            for config_param in config:
                # special parameter kc_user_profile_config
                if config_param in ("kcUserProfileConfig", "kc_user_profile_config"):
                    config_param_org = config_param
                    # rename parameter to be accepted by Keycloak API
                    config_param = "kc.user.profile.config"
                    # make sure no null values are passed to Keycloak API
                    kc_user_profile_config = remove_null_values(config[config_param_org])
                    changeset[camel(component_param)][config_param] = []
                    if len(kc_user_profile_config) > 0:
                        # convert aliases to camelCase
                        kc_user_profile_config = camel_recursive(kc_user_profile_config)
                        # rename validations to be accepted by Keycloak API
                        if "attributes" in kc_user_profile_config[0]:
                            for attribute in kc_user_profile_config[0]["attributes"]:
                                if "validations" in attribute:
                                    if "usernameProhibitedCharacters" in attribute["validations"]:
                                        attribute["validations"]["username-prohibited-characters"] = attribute[
                                            "validations"
                                        ].pop("usernameProhibitedCharacters")
                                    if "upUsernameNotIdnHomograph" in attribute["validations"]:
                                        attribute["validations"]["up-username-not-idn-homograph"] = attribute[
                                            "validations"
                                        ].pop("upUsernameNotIdnHomograph")
                                    if "personNameProhibitedCharacters" in attribute["validations"]:
                                        attribute["validations"]["person-name-prohibited-characters"] = attribute[
                                            "validations"
                                        ].pop("personNameProhibitedCharacters")
                                    if "isoDate" in attribute["validations"]:
                                        attribute["validations"]["iso-date"] = attribute["validations"].pop("isoDate")
                                    if "localDate" in attribute["validations"]:
                                        attribute["validations"]["local-date"] = attribute["validations"].pop(
                                            "localDate"
                                        )
                        changeset[camel(component_param)][config_param].append(kc_user_profile_config[0])
                # usual camelCase parameters
                else:
                    changeset[camel(component_param)][camel(config_param)] = []
                    raw_value = module.params.get(component_param)[config_param]
                    if isinstance(raw_value, bool):
                        value = str(raw_value).lower()
                    else:
                        value = raw_value  # Directly use the raw value
                    changeset[camel(component_param)][camel(config_param)].append(value)
        # usual parameters
        else:
            new_param_value = module.params.get(component_param)
            changeset[camel(component_param)] = new_param_value

    # Make it easier to refer to current module parameters
    state = module.params.get("state")
    parent_id = module.params.get("parent_id")
    provider_type = module.params.get("provider_type")
    provider_id = module.params.get("provider_id")

    # Make a deep copy of the changeset. This is use when determining
    # changes to the current state.
    changeset_copy = deepcopy(changeset)

    # Get a list of all Keycloak components that are of userprofile provider type.
    realm_userprofiles = kc.get_components(urlencode(dict(type=provider_type)), parent_id)

    # If this component is present get its userprofile ID. Confusingly the userprofile ID is
    # also known as the Provider ID.
    userprofile_id = None

    # Track individual parameter changes
    changes = ""

    # This tells Ansible whether the userprofile was changed (added, removed, modified)
    result["changed"] = False

    # Loop through the list of components. If we encounter a component whose
    # name matches the value of the name parameter then assume the userprofile is
    # already present.
    for userprofile in realm_userprofiles:
        if provider_id == "declarative-user-profile":
            userprofile_id = userprofile["id"]
            changeset["id"] = userprofile_id
            changeset_copy["id"] = userprofile_id

            # keycloak returns kc.user.profile.config as a single JSON formatted string, so we have to deserialize it
            if "config" in userprofile and "kc.user.profile.config" in userprofile["config"]:
                userprofile["config"]["kc.user.profile.config"][0] = json.loads(
                    userprofile["config"]["kc.user.profile.config"][0]
                )

            # Compare top-level parameters
            for param in changeset:
                before_realm_userprofile[param] = userprofile[param]

                if changeset_copy[param] != userprofile[param] and param != "config":
                    changes += f"{param}: {userprofile[param]} -> {changeset_copy[param]}, "
                    result["changed"] = True

            # Compare parameters under the "config" userprofile
            for p, v in changeset_copy["config"].items():
                before_realm_userprofile["config"][p] = userprofile["config"][p]
                if v != userprofile["config"][p]:
                    changes += f"config.{p}: {userprofile['config'][p]} -> {v}, "
                    result["changed"] = True

    # Check all the possible states of the resource and do what is needed to
    # converge current state with desired state (create, update or delete
    # the userprofile).

    # keycloak expects kc.user.profile.config as a single JSON formatted string, so we have to serialize it
    if "config" in changeset and "kc.user.profile.config" in changeset["config"]:
        changeset["config"]["kc.user.profile.config"][0] = json.dumps(changeset["config"]["kc.user.profile.config"][0])
    if userprofile_id and state == "present":
        if result["changed"]:
            if module._diff:
                result["diff"] = dict(before=before_realm_userprofile, after=changeset_copy)

            if module.check_mode:
                result["msg"] = f"Userprofile {provider_id} would be changed: {changes.strip(', ')}"
            else:
                kc.update_component(changeset, parent_id)
                result["msg"] = f"Userprofile {provider_id} changed: {changes.strip(', ')}"
        else:
            result["msg"] = f"Userprofile {provider_id} was in sync"

        result["end_state"] = changeset_copy
    elif userprofile_id and state == "absent":
        if module._diff:
            result["diff"] = dict(before=before_realm_userprofile, after={})

        if module.check_mode:
            result["changed"] = True
            result["msg"] = f"Userprofile {provider_id} would be deleted"
        else:
            kc.delete_component(userprofile_id, parent_id)
            result["changed"] = True
            result["msg"] = f"Userprofile {provider_id} deleted"

        result["end_state"] = {}
    elif not userprofile_id and state == "present":
        if module._diff:
            result["diff"] = dict(before={}, after=changeset_copy)

        if module.check_mode:
            result["changed"] = True
            result["msg"] = f"Userprofile {provider_id} would be created"
        else:
            kc.create_component(changeset, parent_id)
            result["changed"] = True
            result["msg"] = f"Userprofile {provider_id} created"

        result["end_state"] = changeset_copy
    elif not userprofile_id and state == "absent":
        result["changed"] = False
        result["msg"] = f"Userprofile {provider_id} not present"
        result["end_state"] = {}

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class DNSConversionError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.error_message = message



#!/usr/bin/python
# Copyright (c) 2017 Chris Hoffman <christopher.hoffman@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: npm
short_description: Manage node.js packages with npm
description:
  - Manage node.js packages with Node Package Manager (npm).
author: "Chris Hoffman (@chrishoffman)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - The name of a node.js library to install.
    type: str
  path:
    description:
      - The base path where to install the node.js libraries.
      - When O(global=true), then O(path) is not used to install the packages.
    type: path
  version:
    description:
      - The version to be installed.
    type: str
  global:
    description:
      - Install the node.js library globally.
      - When O(global=true), then O(path) is not used to install the packages.
    default: false
    type: bool
  executable:
    description:
      - The executable location for npm.
      - This is useful if you are using a version manager, such as nvm.
    type: path
  ignore_scripts:
    description:
      - Use the C(--ignore-scripts) flag when installing.
    type: bool
    default: false
  unsafe_perm:
    description:
      - Use the C(--unsafe-perm) flag when installing.
    type: bool
    default: false
  ci:
    description:
      - Install packages based on package-lock file, same as running C(npm ci).
    type: bool
    default: false
  production:
    description:
      - Install dependencies in production mode, excluding devDependencies.
    type: bool
    default: false
  registry:
    description:
      - The registry to install modules from.
    type: str
  state:
    description:
      - The state of the node.js library.
    type: str
    default: present
    choices: ["present", "absent", "latest"]
  no_optional:
    description:
      - Use the C(--no-optional) flag when installing.
    type: bool
    default: false
    version_added: 2.0.0
  no_bin_links:
    description:
      - Use the C(--no-bin-links) flag when installing.
    type: bool
    default: false
    version_added: 2.5.0
  force:
    description:
      - Use the C(--force) flag when installing.
    type: bool
    default: false
    version_added: 9.5.0
requirements:
  - npm installed in bin path (recommended /usr/local/bin)
"""

EXAMPLES = r"""
- name: Install "coffee-script" node.js package.
  community.general.npm:
    name: coffee-script
    path: /app/location

- name: Install "coffee-script" node.js package on version 1.6.1.
  community.general.npm:
    name: coffee-script
    version: '1.6.1'
    path: /app/location

- name: Install "coffee-script" node.js package globally.
  community.general.npm:
    name: coffee-script
    global: true

- name: Force Install "coffee-script" node.js package.
  community.general.npm:
    name: coffee-script
    force: true

- name: Remove the globally package "coffee-script".
  community.general.npm:
    name: coffee-script
    global: true
    state: absent

- name: Install "coffee-script" node.js package from custom registry.
  community.general.npm:
    name: coffee-script
    registry: 'http://registry.mysite.com'

- name: Install packages based on package.json.
  community.general.npm:
    path: /app/location

- name: Update packages based on package.json to their latest version.
  community.general.npm:
    path: /app/location
    state: latest

- name: Install packages based on package.json using the npm installed with nvm v0.10.1.
  community.general.npm:
    path: /app/location
    executable: /opt/nvm/v0.10.1/bin/npm
    state: present
"""

import json
import os
import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt


class Npm:
    def __init__(self, module, **kwargs):
        self.module = module
        self.glbl = kwargs["glbl"]
        self.name = kwargs["name"]
        self.version = kwargs["version"]
        self.path = kwargs["path"]
        self.registry = kwargs["registry"]
        self.production = kwargs["production"]
        self.ignore_scripts = kwargs["ignore_scripts"]
        self.unsafe_perm = kwargs["unsafe_perm"]
        self.state = kwargs["state"]
        self.no_optional = kwargs["no_optional"]
        self.no_bin_links = kwargs["no_bin_links"]
        self.force = kwargs["force"]

        if kwargs["executable"]:
            self.executable = kwargs["executable"].split(" ")
        else:
            self.executable = [module.get_bin_path("npm", True)]

        if kwargs["version"] and kwargs["state"] != "absent":
            self.name_version = f"{self.name}@{kwargs['version']}"
        else:
            self.name_version = self.name

        self.runner = CmdRunner(
            module,
            command=self.executable,
            arg_formats=dict(
                exec_args=cmd_runner_fmt.as_list(),
                global_=cmd_runner_fmt.as_bool("--global"),
                production=cmd_runner_fmt.as_bool("--production"),
                ignore_scripts=cmd_runner_fmt.as_bool("--ignore-scripts"),
                unsafe_perm=cmd_runner_fmt.as_bool("--unsafe-perm"),
                name_version=cmd_runner_fmt.as_list(),
                registry=cmd_runner_fmt.as_opt_val("--registry"),
                no_optional=cmd_runner_fmt.as_bool("--no-optional"),
                no_bin_links=cmd_runner_fmt.as_bool("--no-bin-links"),
                force=cmd_runner_fmt.as_bool("--force"),
            ),
        )

    def _exec(self, args, run_in_check_mode=False, check_rc=True, add_package_name=True):
        if not self.module.check_mode or (self.module.check_mode and run_in_check_mode):
            # If path is specified, cd into that path and run the command.
            cwd = None
            if self.path:
                if not os.path.exists(self.path):
                    os.makedirs(self.path)
                if not os.path.isdir(self.path):
                    self.module.fail_json(msg=f"path {self.path} is not a directory")
                cwd = self.path

            params = dict(self.module.params)
            params["exec_args"] = args
            params["global_"] = self.glbl
            params["production"] = self.production and ("install" in args or "update" in args or "ci" in args)
            params["name_version"] = self.name_version if add_package_name else None

            with self.runner(
                "exec_args global_ production ignore_scripts unsafe_perm name_version registry no_optional no_bin_links force",
                check_rc=check_rc,
                cwd=cwd,
            ) as ctx:
                rc, out, err = ctx.run(**params)
            return out

        return ""

    def list(self):
        cmd = ["list", "--json", "--long"]

        installed = list()
        missing = list()
        data = {}
        try:
            data = json.loads(self._exec(cmd, True, False, False) or "{}")
        except getattr(json, "JSONDecodeError", ValueError) as e:
            self.module.fail_json(msg=f"Failed to parse NPM output with error {e}")
        if "dependencies" in data:
            for dep, props in data["dependencies"].items():
                if "missing" in props and props["missing"]:
                    missing.append(dep)
                elif "invalid" in props and props["invalid"]:
                    missing.append(dep)
                else:
                    installed.append(dep)
                    if "version" in props and props["version"]:
                        dep_version = f"{dep}@{props['version']}"
                        installed.append(dep_version)
            if self.name_version and self.name_version not in installed:
                missing.append(self.name)
        # Named dependency not installed
        else:
            missing.append(self.name)

        return installed, missing

    def install(self):
        return self._exec(["install"])

    def ci_install(self):
        return self._exec(["ci"])

    def update(self):
        return self._exec(["update"])

    def uninstall(self):
        return self._exec(["uninstall"])

    def list_outdated(self):
        outdated = list()
        data = self._exec(["outdated"], True, False)
        for dep in data.splitlines():
            if dep:
                # node.js v0.10.22 changed the `npm outdated` module separator
                # from "@" to " ". Split on both for backwards compatibility.
                pkg, other = re.split(r"\s|@", dep, maxsplit=1)
                outdated.append(pkg)

        return outdated


def main():
    arg_spec = dict(
        name=dict(type="str"),
        path=dict(type="path"),
        version=dict(type="str"),
        production=dict(default=False, type="bool"),
        executable=dict(type="path"),
        registry=dict(type="str"),
        state=dict(default="present", choices=["present", "absent", "latest"]),
        ignore_scripts=dict(default=False, type="bool"),
        unsafe_perm=dict(default=False, type="bool"),
        ci=dict(default=False, type="bool"),
        no_optional=dict(default=False, type="bool"),
        no_bin_links=dict(default=False, type="bool"),
        force=dict(default=False, type="bool"),
    )
    arg_spec["global"] = dict(default=False, type="bool")
    module = AnsibleModule(
        argument_spec=arg_spec,
        required_if=[
            ("state", "absent", ["name"]),
            ("global", False, ["path"]),
        ],
        supports_check_mode=True,
    )

    name = module.params["name"]
    path = module.params["path"]
    version = module.params["version"]
    glbl = module.params["global"]
    state = module.params["state"]

    npm = Npm(
        module,
        name=name,
        path=path,
        version=version,
        glbl=glbl,
        production=module.params["production"],
        executable=module.params["executable"],
        registry=module.params["registry"],
        ignore_scripts=module.params["ignore_scripts"],
        unsafe_perm=module.params["unsafe_perm"],
        state=state,
        no_optional=module.params["no_optional"],
        no_bin_links=module.params["no_bin_links"],
        force=module.params["force"],
    )

    changed = False
    if module.params["ci"]:
        npm.ci_install()
        changed = True
    elif state == "present":
        installed, missing = npm.list()
        if missing:
            changed = True
            npm.install()
    elif state == "latest":
        installed, missing = npm.list()
        outdated = npm.list_outdated()
        if missing:
            changed = True
            npm.install()
        if outdated:
            changed = True
            npm.update()
    else:  # absent
        installed, missing = npm.list()
        if name in installed:
            changed = True
            npm.uninstall()

    module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

"""
The vlan_expander plugin code
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type


def vlan_expander(data):
    expanded_list = []
    for each in data.split(","):
        if "-" in each:
            f, t = map(int, each.split("-"))
            expanded_list.extend(range(f, t + 1))
        else:
            expanded_list.append(int(each))
    return sorted(expanded_list)



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ml_environment_info
version_added: "3.16.0"
short_description: List ML Environment resources
description:
    - List ML Environment resources.
options:
    name:
        description:
            - Name of the Azure ML environment.
        type: str
        required: false
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    ml_workspace:
        description:
            - Name of the Azure ML workspace.
        required: true
        type: str
    ml_registry:
        description:
            - If provided, the command will target the registry
              instead of a workspace.  Hence resource group and
              workspace won't be required. Must be provided if
              ml_workspace and resource_group are not provided.
        required: false
        type: str
    version:
        description:
            - Version of the data asset. Must be provided, if label
              is not provided. Mutually exclusive with label.
        required: false
        type: str
    label:
        description:
            - Label of the data asset. Must be provided, if version
              is not provided. Mutually exclusive with version.
        required: false
        type: str
    list_type:
        description:
            - Specify if you want active, archived or all datasets.
        type: str
        choices:
            - active
            - archived
            - all

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: List ML Environment by name
  azure.azcollection.azure_rm_ml_environment_info:
    name: MyEnvironment
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace

- name: List Archived ML Environments
  azure.azcollection.azure_rm_ml_environment_info:
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace
    list_type: archived
'''

RETURN = '''
ml_environments:
    description:
        - Environments that match the query.
    returned: always
    type: dict
    sample: [
      {
          "conda_file": {
              "channels": [
                  "conda-forge"
              ],
              "dependencies": [
                  "python=3.8",
                  "pip=20.0.2",
                  {
                      "pip": [
                          "numpy",
                          "pandas",
                          "pyarrow",
                          "scikit-learn",
                          "mlflow",
                          "azureml-mlflow",
                          "responsibleai~=0.17.0",
                          "raiwidgets~=0.17.0",
                          "markupsafe<=2.0.1",
                          "itsdangerous==2.0.1",
                          "azureml-dataset-runtime",
                          "azureml-core"
                      ]
                  }
              ],
              "name": "aml-rai-environment"
          },
          "creation_context": {
              "created_at": "2026-03-20T17:34:28.790877+00:00",
              "created_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
              "created_by_type": "Application",
              "last_modified_at": "2026-03-20T17:34:28.790877+00:00",
              "last_modified_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
              "last_modified_by_type": "Application"
          },
          "id": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-environment/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/environments/environmentxxxxxxxxxx/versions/2",
          "image": "mcr.microsoft.com/azureml/openmpi3.1.2-ubuntu18.04:20220303.v1",
          "name": "environmentxxxxxxxxxx",
          "os_type": "linux",
          "tags": {},
          "version": "2"
      }
    ]
'''  # NOQA


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ml import MLClientCommon
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMMLEnvironmentInfo(MLClientCommon):
    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=False,
            ),
            resource_group=dict(
                type='str',
                required=True,
            ),
            ml_workspace=dict(
                type='str',
                required=True,
            ),
            ml_registry=dict(
                type='str',
            ),
            version=dict(
                type='str',
            ),
            label=dict(
                type='str',
            ),
            list_type=dict(
                type='str',
                choices=['archived', 'active', 'all'],
            ),
        )

        self._client = None

        self.results = dict(
            ml_environments=[]
        )

        required_one_of = [('ml_workspace', 'ml_registry')]
        mutually_exclusive = [('version', 'label'),
                              ('ml_workspace', 'ml_registry'),
                              ('name', 'list_type')]
        required_together = [('ml_workspace', 'resource_group')]

        super(AzureRMMLEnvironmentInfo, self).__init__(self.module_arg_spec,
                                                       supports_tags=False,
                                                       supports_check_mode=True,
                                                       facts_module=True,
                                                       required_one_of=required_one_of,
                                                       mutually_exclusive=mutually_exclusive,
                                                       required_together=required_together
                                                       )

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        if self.name:
            try:
                result = self.client.environments.get(name=self.name,
                                                      version=self.version,
                                                      label=self.label)
                ml_environments = [self.entity_to_dict(result)]
            except ResourceNotFoundError:
                ml_environments = []
        else:
            list_view_type = self.get_list_view_type(self.list_type)
            results = self.client.environments.list(name=self.name,
                                                    list_view_type=list_view_type)
            ml_environments = [self.entity_to_dict(self.client.environments.get(x.name, version=x.latest_version)) for x in results]

        self.results['ml_environments'] = ml_environments
        return self.results


def main():
    AzureRMMLEnvironmentInfo()


if __name__ == '__main__':
    main()



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025 Felix Fontein <felix@fontein.de>

# /// script
# dependencies = ["nox>=2025.02.09", "antsibull-nox"]
# ///

import os
import sys

import nox  # type: ignore[import-not-found]

# Whether the noxfile is running in CI:
IN_CI = os.environ.get("CI") == "true"


try:
    import antsibull_nox  # type: ignore[import-not-found]
except ImportError:
    print("You need to install antsibull-nox in the same Python environment as nox.")
    sys.exit(1)


antsibull_nox.load_antsibull_nox_toml()


@nox.session(name="aliases", python=False, default=True)
def aliases(session: nox.Session) -> None:
    session.run("python", "tests/sanity/extra/aliases.py")


@nox.session(name="botmeta", default=True)
def botmeta(session: nox.Session) -> None:
    session.install("PyYAML", "voluptuous")
    session.run("python", "tests/sanity/extra/botmeta.py")


@nox.session(name="ansible-output", default=False)
def ansible_output(session: nox.Session) -> None:
    session.install(
        "ansible-core",
        "antsibull-docs",
        # Needed libs for some code blocks:
        "jc",
        "hashids",
        # Tools for post-processing
        "ruamel.yaml",  # used by docs/docsite/reformat-yaml.py
    )
    args = []
    if IN_CI:
        args.append("--check")
    session.run("antsibull-docs", "ansible-output", *args, *session.posargs)


# Allow to run the noxfile with `python noxfile.py`, `pipx run noxfile.py`, or similar.
# Requires nox >= 2025.02.09
if __name__ == "__main__":
    nox.main()



# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.dns.plugins.module_utils._conversion.base import (
    DNSConversionError,
)
from ansible_collections.community.dns.plugins.module_utils._conversion.txt import (
    decode_txt_value,
    encode_txt_value,
)
from ansible_collections.community.dns.plugins.module_utils._record import DNSRecord

if t.TYPE_CHECKING:
    from collections.abc import Sequence  # pragma: no cover

    from .._argspec import OptionProvider  # pragma: no cover
    from .._provider import ProviderInformation  # pragma: no cover
    from .._record import RecordIDT  # pragma: no cover
    from .._record_set import DNSRecordSet, RecordSetIDT  # pragma: no cover


class RecordConverter:
    def __init__(
        self, provider_information: ProviderInformation, option_provider: OptionProvider
    ) -> None:
        """
        Create a record converter.
        """
        self._provider_information = provider_information
        self._option_provider = option_provider

        # Valid values: 'decoded', 'encoded', 'encoded-no-char-encoding'
        self._txt_api_handling = self._provider_information.txt_record_handling()
        self._txt_api_character_encoding = (
            self._provider_information.txt_character_encoding()
        )
        self._txt_always_quote = self._provider_information.txt_always_quote()
        # Valid values: 'api', 'quoted', 'unquoted'
        self._txt_transformation = self._option_provider.get_option(
            "txt_transformation"
        )
        # Valid values: 'decimal', 'octal'
        self._txt_character_encoding = self._option_provider.get_option(
            "txt_character_encoding"
        )

    def emit_deprecations(self, deprecator: t.Callable[[str], None]) -> None:
        pass

    def _handle_txt_api(self, to_api: bool, record: DNSRecord[RecordIDT]) -> None:
        """
        Handle TXT records for sending to/from the API.
        """
        if self._txt_transformation == "api":
            # Do not touch record values
            return

        # We assume that records internally use decoded values
        if self._txt_api_handling in (
            "encoded",
            "encoded-no-char-encoding",
        ):
            if to_api:
                record.target = encode_txt_value(
                    record.target,
                    always_quote=self._txt_always_quote,
                    use_character_encoding=self._txt_api_handling == "encoded",
                    character_encoding=self._txt_api_character_encoding,
                )
            else:
                record.target = decode_txt_value(
                    record.target, character_encoding=self._txt_api_character_encoding
                )

    def _handle_txt_user(self, to_user: bool, record: DNSRecord[RecordIDT]) -> None:
        """
        Handle TXT records for sending to/from the user.
        """
        if self._txt_transformation == "api":
            # Do not touch record values
            return

        # We assume that records internally use decoded values
        if self._txt_transformation == "quoted":
            if to_user:
                record.target = encode_txt_value(
                    record.target, character_encoding=self._txt_character_encoding
                )
            else:
                record.target = decode_txt_value(
                    record.target, character_encoding=self._txt_character_encoding
                )

    def process_from_api(self, record: DNSRecord[RecordIDT]) -> DNSRecord[RecordIDT]:
        """
        Process a record object (DNSRecord) after receiving from API.
        Modifies the record in-place.
        """
        try:
            record.target = to_text(record.target)
            if record.type == "TXT":
                self._handle_txt_api(False, record)
            return record
        except DNSConversionError as e:
            raise DNSConversionError(
                f"While processing record from API: {e.error_message}"
            ) from e

    def process_to_api(self, record: DNSRecord[RecordIDT]) -> DNSRecord[RecordIDT]:
        """
        Process a record object (DNSRecord) for sending to API.
        Modifies the record in-place.
        """
        try:
            if record.type == "TXT":
                self._handle_txt_api(True, record)
            return record
        except DNSConversionError as e:  # pragma: no cover
            # This can never happen
            raise DNSConversionError(
                f"While processing record for the API: {e.error_message}"
            ) from e  # pragma: no cover

    def process_from_user(self, record: DNSRecord[RecordIDT]) -> DNSRecord[RecordIDT]:
        """
        Process a record object (DNSRecord) after receiving from the user.
        Modifies the record in-place.
        """
        try:
            record.target = to_text(record.target)
            if record.type == "TXT":
                self._handle_txt_user(False, record)
            return record
        except DNSConversionError as e:
            raise DNSConversionError(
                f"While processing record from the user: {e.error_message}"
            ) from e

    def process_to_user(self, record: DNSRecord[RecordIDT]) -> DNSRecord[RecordIDT]:
        """
        Process a record object (DNSRecord) for sending to the user.
        Modifies the record in-place.
        """
        try:
            if record.type == "TXT":
                self._handle_txt_user(True, record)
            return record
        except DNSConversionError as e:  # pragma: no cover
            # This can never happen
            raise DNSConversionError(
                f"While processing record for the user: {e.error_message}"
            ) from e  # pragma: no cover

    def clone_from_api(self, record: DNSRecord[RecordIDT]) -> DNSRecord[RecordIDT]:
        """
        Process a record object (DNSRecord) after receiving from API.
        Return a modified clone of the record; the original will not be modified.
        """
        record = record.clone()
        self.process_from_api(record)
        return record

    def clone_to_api(self, record: DNSRecord[RecordIDT]) -> DNSRecord[RecordIDT]:
        """
        Process a record object (DNSRecord) for sending to API.
        Return a modified clone of the record; the original will not be modified.
        """
        record = record.clone()
        self.process_to_api(record)
        return record

    def clone_multiple_from_api(
        self, records: Sequence[DNSRecord[RecordIDT]]
    ) -> list[DNSRecord[RecordIDT]]:
        """
        Process a list of record object (DNSRecord) after receiving from API.
        Return a list of modified clones of the records; the originals will not be modified.
        """
        return [self.clone_from_api(record) for record in records]

    def clone_multiple_to_api(
        self, records: Sequence[DNSRecord[RecordIDT]]
    ) -> list[DNSRecord[RecordIDT]]:
        """
        Process a list of record objects (DNSRecord) for sending to API.
        Return a list of modified clones of the records; the originals will not be modified.
        """
        return [self.clone_to_api(record) for record in records]

    def clone_set_to_api(
        self, record_set: DNSRecordSet[RecordSetIDT, RecordIDT]
    ) -> DNSRecordSet[RecordSetIDT, RecordIDT]:
        """
        Process a record set object (DNSRecordSet) for sending to API.
        Return a modified clone of the record set; the original will not be modified.
        """
        record_set = record_set.clone()
        record_set.records = [
            self.clone_to_api(record) for record in record_set.records
        ]
        return record_set

    def process_multiple_from_api(
        self, records: list[DNSRecord[RecordIDT]]
    ) -> list[DNSRecord[RecordIDT]]:
        """
        Process a list of record object (DNSRecord) after receiving from API.
        Modifies the records in-place.
        """
        for record in records:
            self.process_from_api(record)
        return records

    def process_multiple_to_api(
        self, records: list[DNSRecord[RecordIDT]]
    ) -> list[DNSRecord[RecordIDT]]:
        """
        Process a list of record objects (DNSRecord) for sending to API.
        Modifies the records in-place.
        """
        for record in records:
            self.process_to_api(record)
        return records

    def process_multiple_from_user(
        self, records: list[DNSRecord[RecordIDT]]
    ) -> list[DNSRecord[RecordIDT]]:
        """
        Process a list of record object (DNSRecord) after receiving from the user.
        Modifies the records in-place.
        """
        for record in records:
            self.process_from_user(record)
        return records

    def process_multiple_to_user(
        self, records: list[DNSRecord[RecordIDT]]
    ) -> list[DNSRecord[RecordIDT]]:
        """
        Process a list of record objects (DNSRecord) for sending to the user.
        Modifies the records in-place.
        """
        for record in records:
            self.process_to_user(record)
        return records

    def process_set_from_api(
        self, record_set: DNSRecordSet[RecordSetIDT, RecordIDT]
    ) -> DNSRecordSet[RecordSetIDT, RecordIDT]:
        """
        Process a record set object (DNSRecordSet) after receiving from API.
        Modifies the records in-place.
        """
        for record in record_set.records:
            self.process_from_api(record)
        return record_set

    def process_set_to_api(
        self, record_set: DNSRecordSet[RecordSetIDT, RecordIDT]
    ) -> DNSRecordSet[RecordSetIDT, RecordIDT]:
        """
        Process a record set object (DNSRecordSet) for sending to API.
        Modifies the records in-place.
        """
        for record in record_set.records:
            self.process_to_api(record)
        return record_set

    def process_set_from_user(
        self, record_set: DNSRecordSet[RecordSetIDT, RecordIDT]
    ) -> DNSRecordSet[RecordSetIDT, RecordIDT]:
        """
        Process a record set object (DNSRecordSet) after receiving from the user.
        Modifies the records in-place.
        """
        for record in record_set.records:
            self.process_from_user(record)
        return record_set

    def process_set_to_user(
        self, record_set: DNSRecordSet[RecordSetIDT, RecordIDT]
    ) -> DNSRecordSet[RecordSetIDT, RecordIDT]:
        """
        Process a record set objects (DNSRecordSet) for sending to the user.
        Modifies the records in-place.
        """
        for record in record_set.records:
            self.process_to_user(record)
        return record_set

    def process_value_from_user(self, record_type: str, value: str) -> str:
        """
        Process a record value (string) after receiving from the user.
        """
        record: DNSRecord[None] = DNSRecord(
            record_id=None, record_type=record_type, target=value
        )
        self.process_from_user(record)
        return record.target

    def process_values_from_user(
        self, record_type: str, values: Sequence[str]
    ) -> list[str]:
        """
        Process a list of record values (strings) after receiving from the user.
        """
        return [self.process_value_from_user(record_type, value) for value in values]

    def process_value_to_user(self, record_type: str, value: str) -> str:
        """
        Process a record value (string) for sending to the user.
        """
        record: DNSRecord[None] = DNSRecord(
            record_id=None, record_type=record_type, target=value
        )
        self.process_to_user(record)
        return record.target

    def process_values_to_user(
        self, record_type: str, values: Sequence[str]
    ) -> list[str]:
        """
        Process a list of record values (strings) for sending to the user.
        """
        return [self.process_value_to_user(record_type, value) for value in values]



FILTER_KEY_MAP = {
    "both-way": "filterRelationships",
    "consumer-to-provider": "filterRelationshipsConsumerToProvider",
    "provider-to-consumer": "filterRelationshipsProviderToConsumer",
}

PRIORITY_MAP = {
    "default": "default",
    "lowest_priority": "level1",
    "medium_priority": "level2",
    "highest_priority": "level3",
}

SERVICE_NODE_CONNECTOR_MAP = {
    "bd": {"id": "bd", "connector_type": "general"}
    # 'external_epg': {'id': 'externalEpg', 'connector_type': 'route-peering'}
}

YES_OR_NO_TO_BOOL_STRING_MAP = {"yes": "true", "no": "false", True: "yes", False: "no"}

ENABLED_OR_DISABLED_TO_BOOL_STRING_MAP = {"enabled": True, "disabled": False}
BOOL_TO_ENABLED_OR_DISABLED_STRING_MAP = {True: "enabled", False: "disabled"}

NDO_4_UNIQUE_IDENTIFIERS = ["templateID", "autoRouteTargetImport", "autoRouteTargetExport"]

NDO_API_VERSION_FORMAT = "/mso/api/{api_version}"
NDO_API_VERSION_PATH_FORMAT = "/mso/api/{api_version}/{path}"

NDO_CIPHER_SUITE_MAP = {
    "128_gcm_aes": "128GcmAes",
    "128_gcm_aes_xpn": "128GcmAesXpn",
    "256_gcm_aes": "256GcmAes",
    "256_gcm_aes_xpn": "256GcmAesXpn",
}

NDO_SECURITY_POLICY_MAP = {
    "should_secure": "shouldSecure",
    "must_secure": "mustSecure",
}

EPG_U_SEG_ATTR_TYPE_MAP = {
    "ip": "ip",
    "mac": "mac",
    "dns": "dns",
    "vm_datacenter": "rootContName",
    "vm_hypervisor_identifier": "hv",
    "vm_operating_system": "guest-os",
    "vm_tag": "tag",
    "vm_identifier": "vm",
    "vmm_domain": "domain",
    "vm_name": "vm-name",
    "vnic_dn": "vnic",
}

EPG_U_SEG_ATTR_OPERATOR_LIST = ["equals", "contains", "starts_with", "ends_with"]

AZURE_L4L7_CONNECTOR_TYPE_MAP = {
    "none": "none",
    "redirect": "redir",
    "source_nat": "snat",
    "destination_nat": "dnat",
    "source_and_destination_nat": "snat_dnat",
}

LISTENER_PROTOCOLS = ["http", "https", "tcp", "udp", "tls", "inherit"]

LISTENER_SECURITY_POLICY_MAP = {
    "default": "default",
    "elb_sec_2016_18": "eLBSecurityPolicy-2016-08",
    "elb_sec_fs_2018_06": "eLBSecurityPolicy-FS-2018-06",
    "elb_sec_tls_1_2_2017_01": "eLBSecurityPolicy-TLS-1-2-2017-01",
    "elb_sec_tls_1_2_ext_2018_06": "eLBSecurityPolicy-TLS-1-2-Ext-2018-06",
    "elb_sec_tls_1_1_2017_01": "eLBSecurityPolicy-TLS-1-1-2017-01",
    "elb_sec_2015_05": "eLBSecurityPolicy-2015-05",
    "elb_sec_tls_1_0_2015_04": "eLBSecurityPolicy-TLS-1-0-2015-04",
    "app_gw_ssl_default": "AppGwSslPolicyDefault",
    "app_gw_ssl_2015_501": "AppGwSslPolicy20150501",
    "app_gw_ssl_2017_401": "AppGwSslPolicy20170401",
    "app_gw_ssl_2017_401s": "AppGwSslPolicy20170401S",
}

LISTENER_ACTION_TYPE_MAP = {"fixed_response": "fixedResponse", "forward": "forward", "redirect": "redirect", "ha_port": "haPort"}

LISTENER_CONTENT_TYPE_MAP = {"text_plain": "textPlain", "text_css": "textCSS", "text_html": "textHtml", "app_js": "appJS", "app_json": "appJson"}

LISTENER_REDIRECT_CODE_MAP = {
    "unknown": "unknown",
    "permanently_moved": "permMoved",
    "found": "found",
    "see_other": "seeOther",
    "temporary_redirect": "temporary",
}

PORT_CHANNEL_MODE_MAP = {
    "lacp_active": "active",
    "lacp_passive": "passive",
    "static_channel_mode_on": "off",
    "mac_pinning": "mac-pin",
    "mac_pinning_physical_nic_load": "mac-pin-nicload",
    "use_explicit_failover_order": "explicit-failover",
}

CONTROL_MAP = {
    "fast_sel_hot_stdby": "fast-sel-hot-stdby",
    "graceful_conv": "graceful-conv",
    "susp_individual": "susp-individual",
    "load_defer": "load-defer",
    "symmetric_hash": "symmetric-hash",
}

LINK_LEVEL_FEC_MAP = {
    "inherit": "inherit",
    "cl74_fc_fec": "cl74-fc-fec",
    "cl91_rs_fec": "cl91-rs-fec",
    "cons16_rs_fec": "cons16-rs-fec",
    "ieee_rs_fec": "ieee-rs-fec",
    "kp_fec": "kp-fec",
    "disable_fec": "disable-fec",
}

L2_INTERFACE_QINQ_MAP = {
    "double_q_tag_port": "doubleQtagPort",
    "core_port": "corePort",
    "edge_port": "edgePort",
    "disabled": "disabled",
}

LOAD_BALANCE_HASHING_MAP = {
    "destination_ip": "dst-ip",
    "layer_4_destination_ip": "l4-dst-port",
    "layer_4_source_ip": "l4-src-port",
    "source_ip": "src-ip",
}

TEMPLATE_TYPES = {
    "tenant": {
        "template_type": "tenantPolicy",  # templateType in payload
        "template_type_container": "tenantPolicyTemplate",  # templateType container in payload
        "tenant": True,  # tenant required
        "site_amount": 2,  # 1 = 1 site, 2 = multiple sites
        "template_container": True,  # configuration is set in template container in payload
    },
    "l3out": {
        "template_type": "l3out",
        "template_type_container": "l3outTemplate",
        "tenant": True,
        "site_amount": 1,
        "template_container": False,
    },
    "fabric_policy": {
        "template_type": "fabricPolicy",
        "template_type_container": "fabricPolicyTemplate",
        "tenant": False,
        "site_amount": 2,
        "template_container": True,
    },
    "fabric_resource": {
        "template_type": "fabricResource",
        "template_type_container": "fabricResourceTemplate",
        "tenant": False,
        "site_amount": 2,
        "template_container": True,
    },
    "monitoring_tenant": {
        "template_type": "monitoring",
        "template_type_container": "monitoringTemplate",
        "tenant": True,
        "site_amount": 1,
        "template_container": True,
    },
    "monitoring_access": {
        "template_type": "monitoring",
        "template_type_container": "monitoringTemplate",
        "tenant": False,
        "site_amount": 1,
        "template_container": True,
    },
    "service_device": {
        "template_type": "serviceDevice",
        "template_type_container": "deviceTemplate",
        "tenant": True,
        "site_amount": 2,
        "template_container": True,
    },
    "application": {
        "template_type": "application",
        "template_type_container": "appTemplate",
        "tenant": True,
        "site_amount": 2,
        "template_container": True,
    },
}

TARGET_DSCP_MAP = {
    "af11": "af11",
    "af12": "af12",
    "af13": "af13",
    "af21": "af21",
    "af22": "af22",
    "af23": "af23",
    "af31": "af31",
    "af32": "af32",
    "af33": "af33",
    "af41": "af41",
    "af42": "af42",
    "af43": "af43",
    "cs0": "cs0",
    "cs1": "cs1",
    "cs2": "cs2",
    "cs3": "cs3",
    "cs4": "cs4",
    "cs5": "cs5",
    "cs6": "cs6",
    "cs7": "cs7",
    "expedited_forwarding": "expeditedForwarding",
    "voice_admit": "voiceAdmit",
    "unspecified": "unspecified",
}

ORIGINATE_DEFAULT_ROUTE = {"only": "only", "in_addition": "inAddition", "": ""}
L3OUT_ROUTING_PROTOCOLS = {"bgp": ["bgp"], "ospf": ["ospf"], "bgpOspf": ["bgp", "ospf"], None: [None], "": None, "bgpospf": "bgpOspf", "ospfbgp": "bgpOspf"}

QOS_LEVEL = ["unspecified", "level1", "level2", "level3", "level4", "level5", "level6"]
SYNC_E_QUALITY_LEVEL_OPTION = {"option_1": "op1", "option_2_generation_1": "op2g1", "option_2_generation_2": "op2g2"}
PROFILE_TEMPLATE = {"aes67_2015": "aes67", "default": "default", "smpte_2059_2": "smpte", "telecom_8275_1": "telecomFullPath"}

TARGET_COS_MAP = {
    "background": "cos0",
    "best_effort": "cos1",
    "excellent_effort": "cos2",
    "critical_applications": "cos3",
    "video": "cos4",
    "voice": "cos5",
    "internetwork_control": "cos6",
    "network_control": "cos7",
    "unspecified": "cos8",
}

DSCP_COS_KEY_MAP = {
    "dscp_from": "dscpFrom",
    "dscp_to": "dscpTo",
    "dot1p_from": "dot1pFrom",
    "dot1p_to": "dot1pTo",
    "dscp_target": "dscpTarget",
    "target_cos": "targetCos",
    "qos_priority": "priority",
}

LOCAL_ASN_CONFIG = {"none": "none", "no_prepend": "no-prepend", "dual_as": "dual-as", "replace_as": "replace-as"}

QOS_CONGESTION_ALGORITHM_MAP = {"tail_drop": "tailDrop", "wred": "wred"}
QOS_SCHEDULING_ALGORITHM_MAP = {"strict_priority": "strictPriority", "weighted_round_robin": "wrr"}
QOS_PFC_SCOPE_MAP = {"fabric_wide": "fabricWide", "intra_tor": "intraTor"}
COS_VALUES = ["cos0", "cos1", "cos2", "cos3", "cos4", "cos5", "cos6", "cos7", "unspecified"]

PORT_MAPPING = {
    "dns": "53",
    "ftp_data": "20",
    "http": "80",
    "https": "443",
    "pop3": "110",
    "rtsp": "554",
    "smtp": "25",
    "ssh": "22",
    "unspecified": "0",
}

IP_PROTOCOL_MAPPING = {
    "unspecified": "0",
    "egp": "8",
    "eigrp": "88",
    "icmp": "1",
    "icmpv6": "58",
    "igmp": "2",
    "igp": "9",
    "l2tp": "115",
    "ospfigp": "89",
    "pim": "103",
    "tcp": "6",
    "udp": "17",
}

CONTRACT_SERVICE_CHAIN_NODE_FILTER_MAP = {"allow_all": "allow-all", "filters_from_contract": "filters-from-contract"}

PTP_MODES = {"multicast_dynamic": "multicast", "multicast_master": "multicastMaster", "unicast_master": "unicastMaster"}

DOMAIN_TYPE_MAP = {"vmm": "vmmDomain", "physical": "physicalDomain"}

VM_DOMAIN_PROVIDER_MAP = {
    "cloudfoundry": "CloudFoundry",
    "kubernetes": "Kubernetes",
    "microsoft": "Microsoft",
    "openshift": "OpenShift",
    "openstack": "OpenStack",
    "redhat": "Redhat",
    "vmware": "VMware",
    "nutanix": "Nutanix",
}

MATCH_COMMUNITY_SCOPE_MAP = {"transitive": "transitive", "non_transitive": "non-transitive"}

ROUTE_MAP_METRIC_TYPE_MAP = {
    "type1": "ospf-type1",
    "ospf_type1": "ospf-type1",
    "type2": "ospf-type2",
    "ospf_type2": "ospf-type2",
    "": "",
}

ROUTE_MAP_POLICY_MATCH_TYPE = {"prefix_and_routing_policy": "combinable", "routing_policy_only": "global"}

MATCH_PARAMETER_MAP = {
    "destination_ip": "dstIP",
    "destination_ipv4": "dstIPv4",
    "destination_ipv6": "dstIPv6",
    "destination_mac": "dstMac",
    "destination_port": "dstPort",
    "ethertype": "ethertype",
    "ip_protocol": "proto",
    "source_ip": "srcIP",
    "source_ipv4": "srcIPv4",
    "source_ipv6": "srcIPv6",
    "source_mac": "srcMac",
    "source_port": "srcPort",
}



# Copyright 2016 Red Hat | Ansible
# Copyright (c) 2022 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import os
import re
import typing as t
from collections.abc import Mapping, Sequence

from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible.module_utils.parsing.convert_bool import BOOLEANS_FALSE, BOOLEANS_TRUE

from ansible_collections.community.docker.plugins.module_utils._version import (
    LooseVersion,
)

try:
    from requests.exceptions import (  # noqa: F401, pylint: disable=unused-import
        RequestException,
        SSLError,
    )
except ImportError:
    # Define an exception class RequestException so that our code does not break.
    class RequestException(Exception):  # type: ignore
        pass


from ansible_collections.community.docker.plugins.module_utils._api import auth
from ansible_collections.community.docker.plugins.module_utils._api.api.client import (
    APIClient as Client,
)
from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    APIError,
    MissingRequirementException,
    NotFound,
    TLSParameterError,
)
from ansible_collections.community.docker.plugins.module_utils._api.tls import TLSConfig
from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    convert_filters,
    parse_repository_tag,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DEFAULT_DOCKER_HOST,
    DEFAULT_TIMEOUT_SECONDS,
    DEFAULT_TLS,
    DEFAULT_TLS_VERIFY,
    DOCKER_COMMON_ARGS,
    DOCKER_MUTUALLY_EXCLUSIVE,
    DOCKER_REQUIRED_TOGETHER,
    sanitize_result,
    update_tls_hostname,
)

if t.TYPE_CHECKING:
    from collections.abc import Callable


def _get_tls_config(
    fail_function: Callable[[str], t.NoReturn], **kwargs: t.Any
) -> TLSConfig:
    try:
        return TLSConfig(**kwargs)
    except TLSParameterError as exc:
        fail_function(f"TLS config error: {exc}")


def is_using_tls(auth_data: dict[str, t.Any]) -> bool:
    return auth_data["tls_verify"] or auth_data["tls"]


def get_connect_params(
    auth_data: dict[str, t.Any], fail_function: Callable[[str], t.NoReturn]
) -> dict[str, t.Any]:
    if is_using_tls(auth_data):
        auth_data["docker_host"] = auth_data["docker_host"].replace(
            "tcp://", "https://"
        )

    result = {
        "base_url": auth_data["docker_host"],
        "version": auth_data["api_version"],
        "timeout": auth_data["timeout"],
    }

    if auth_data["tls_verify"]:
        # TLS with verification
        tls_config = {
            "verify": True,
            "assert_hostname": auth_data["tls_hostname"],
            "fail_function": fail_function,
        }
        if auth_data["cert_path"] and auth_data["key_path"]:
            tls_config["client_cert"] = (auth_data["cert_path"], auth_data["key_path"])
        if auth_data["cacert_path"]:
            tls_config["ca_cert"] = auth_data["cacert_path"]
        result["tls"] = _get_tls_config(**tls_config)
    elif auth_data["tls"]:
        # TLS without verification
        tls_config = {
            "verify": False,
            "fail_function": fail_function,
        }
        if auth_data["cert_path"] and auth_data["key_path"]:
            tls_config["client_cert"] = (auth_data["cert_path"], auth_data["key_path"])
        result["tls"] = _get_tls_config(**tls_config)

    if auth_data.get("use_ssh_client"):
        result["use_ssh_client"] = True

    # No TLS
    return result


class AnsibleDockerClientBase(Client):
    def __init__(self, min_docker_api_version: str | None = None) -> None:
        self._connect_params = get_connect_params(
            self.auth_params, fail_function=self.fail
        )

        try:
            super().__init__(**self._connect_params)
            self.docker_api_version_str = self.api_version
        except MissingRequirementException as exc:
            self.fail(
                missing_required_lib(exc.requirement), exception=exc.import_exception
            )
        except APIError as exc:
            self.fail(f"Docker API error: {exc}")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error connecting: {exc}")

        self.docker_api_version = LooseVersion(self.docker_api_version_str)
        min_docker_api_version = min_docker_api_version or "1.25"
        if self.docker_api_version < LooseVersion(min_docker_api_version):
            self.fail(
                f"Docker API version is {self.docker_api_version_str}. Minimum version required is {min_docker_api_version}."
            )

    def log(self, msg: t.Any, pretty_print: bool = False) -> None:
        pass
        # if self.debug:
        #     from .util import log_debug
        #     log_debug(msg, pretty_print=pretty_print)

    @abc.abstractmethod
    def fail(self, msg: str, **kwargs: t.Any) -> t.NoReturn:
        pass

    @abc.abstractmethod
    def deprecate(
        self,
        msg: str,
        version: str | None = None,
        date: str | None = None,
        collection_name: str | None = None,
    ) -> None:
        pass

    @staticmethod
    def _get_value(
        param_name: str,
        param_value: t.Any,
        env_variable: str | None,
        default_value: t.Any | None,
        value_type: t.Literal["str", "bool", "int"] = "str",
    ) -> t.Any:
        if param_value is not None:
            # take module parameter value
            if value_type == "bool":
                if param_value in BOOLEANS_TRUE:
                    return True
                if param_value in BOOLEANS_FALSE:
                    return False
                return bool(param_value)
            if value_type == "int":
                return int(param_value)
            return param_value

        if env_variable is not None:
            env_value = os.environ.get(env_variable)
            if env_value is not None:
                # take the env variable value
                if param_name == "cert_path":
                    return os.path.join(env_value, "cert.pem")
                if param_name == "cacert_path":
                    return os.path.join(env_value, "ca.pem")
                if param_name == "key_path":
                    return os.path.join(env_value, "key.pem")
                if value_type == "bool":
                    if env_value in BOOLEANS_TRUE:
                        return True
                    if env_value in BOOLEANS_FALSE:
                        return False
                    return bool(env_value)
                if value_type == "int":
                    return int(env_value)
                return env_value

        # take the default
        return default_value

    @abc.abstractmethod
    def _get_params(self) -> dict[str, t.Any]:
        pass

    @property
    def auth_params(self) -> dict[str, t.Any]:
        # Get authentication credentials.
        # Precedence: module parameters-> environment variables-> defaults.

        self.log("Getting credentials")

        client_params = self._get_params()

        params = {}
        for key in DOCKER_COMMON_ARGS:
            params[key] = client_params.get(key)

        result = {
            "docker_host": self._get_value(
                "docker_host",
                params["docker_host"],
                "DOCKER_HOST",
                DEFAULT_DOCKER_HOST,
                value_type="str",
            ),
            "tls_hostname": self._get_value(
                "tls_hostname",
                params["tls_hostname"],
                "DOCKER_TLS_HOSTNAME",
                None,
                value_type="str",
            ),
            "api_version": self._get_value(
                "api_version",
                params["api_version"],
                "DOCKER_API_VERSION",
                "auto",
                value_type="str",
            ),
            "cacert_path": self._get_value(
                "cacert_path",
                params["ca_path"],
                "DOCKER_CERT_PATH",
                None,
                value_type="str",
            ),
            "cert_path": self._get_value(
                "cert_path",
                params["client_cert"],
                "DOCKER_CERT_PATH",
                None,
                value_type="str",
            ),
            "key_path": self._get_value(
                "key_path",
                params["client_key"],
                "DOCKER_CERT_PATH",
                None,
                value_type="str",
            ),
            "tls": self._get_value(
                "tls", params["tls"], "DOCKER_TLS", DEFAULT_TLS, value_type="bool"
            ),
            "tls_verify": self._get_value(
                "validate_certs",
                params["validate_certs"],
                "DOCKER_TLS_VERIFY",
                DEFAULT_TLS_VERIFY,
                value_type="bool",
            ),
            "timeout": self._get_value(
                "timeout",
                params["timeout"],
                "DOCKER_TIMEOUT",
                DEFAULT_TIMEOUT_SECONDS,
                value_type="int",
            ),
            "use_ssh_client": self._get_value(
                "use_ssh_client",
                params["use_ssh_client"],
                None,
                False,
                value_type="bool",
            ),
        }

        def depr(*args: t.Any, **kwargs: t.Any) -> None:
            self.deprecate(*args, **kwargs)

        update_tls_hostname(
            result,
            old_behavior=True,
            deprecate_function=depr,
            uses_tls=is_using_tls(result),
        )

        return result

    def _handle_ssl_error(self, error: Exception) -> t.NoReturn:
        match = re.match(r"hostname.*doesn\'t match (\'.*\')", str(error))
        if match:
            hostname = self.auth_params["tls_hostname"]
            self.fail(
                f"You asked for verification that Docker daemons certificate's hostname matches {hostname}. "
                f"The actual certificate's hostname is {match.group(1)}. Most likely you need to set DOCKER_TLS_HOSTNAME "
                f"or pass `tls_hostname` with a value of {match.group(1)}. You may also use TLS without verification by "
                "setting the `tls` parameter to true."
            )
        self.fail(f"SSL Exception: {error}")

    def get_container_by_id(self, container_id: str) -> dict[str, t.Any] | None:
        try:
            self.log(f"Inspecting container Id {container_id}")
            result = self.get_json("/containers/{0}/json", container_id)
            self.log("Completed container inspection")
            return result
        except NotFound:
            return None
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error inspecting container: {exc}")

    def get_container(self, name: str | None) -> dict[str, t.Any] | None:
        """
        Lookup a container and return the inspection results.
        """
        if name is None:
            return None

        search_name = name
        if not name.startswith("/"):
            search_name = "/" + name

        result = None
        try:
            params = {
                "limit": -1,
                "all": 1,
                "size": 0,
                "trunc_cmd": 0,
            }
            containers = self.get_json("/containers/json", params=params)
            for container in containers:
                self.log(f"testing container: {container['Names']}")
                if (
                    isinstance(container["Names"], list)
                    and search_name in container["Names"]
                ):
                    result = container
                    break
                if container["Id"].startswith(name):
                    result = container
                    break
                if container["Id"] == name:
                    result = container
                    break
        except SSLError as exc:
            self._handle_ssl_error(exc)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error retrieving container list: {exc}")

        if result is None:
            return None

        return self.get_container_by_id(result["Id"])

    def get_network(
        self, name: str | None = None, network_id: str | None = None
    ) -> dict[str, t.Any] | None:
        """
        Lookup a network and return the inspection results.
        """
        if name is None and network_id is None:
            return None

        result = None

        if network_id is None:
            try:
                networks = self.get_json("/networks")
                for network in networks:
                    self.log(f"testing network: {network['Name']}")
                    if name == network["Name"]:
                        result = network
                        break
                    if network["Id"].startswith(name):
                        result = network
                        break
            except SSLError as exc:
                self._handle_ssl_error(exc)
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error retrieving network list: {exc}")

        if result is not None:
            network_id = result["Id"]

        if network_id is not None:
            try:
                self.log(f"Inspecting network Id {network_id}")
                result = self.get_json("/networks/{0}", network_id)
                self.log("Completed network inspection")
            except NotFound:
                return None
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error inspecting network: {exc}")

        return result

    def _image_lookup(self, name: str, tag: str | None) -> list[dict[str, t.Any]]:
        """
        Including a tag in the name parameter sent to the Docker SDK for Python images method
        does not work consistently. Instead, get the result set for name and manually check
        if the tag exists.
        """
        try:
            params: dict[str, t.Any] = {
                "only_ids": 0,
                "all": 0,
            }
            if LooseVersion(self.api_version) < LooseVersion("1.25"):
                # only use "filter" on API 1.24 and under, as it is deprecated
                params["filter"] = name
            else:
                params["filters"] = convert_filters({"reference": name})
            images = self.get_json("/images/json", params=params)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error searching for image {name} - {exc}")
        if tag:
            lookup = f"{name}:{tag}"
            lookup_digest = f"{name}@{tag}"
            response = images
            images = []
            for image in response:
                tags = image.get("RepoTags")
                digests = image.get("RepoDigests")
                if (tags and lookup in tags) or (digests and lookup_digest in digests):
                    images = [image]
                    break
        return images

    def find_image(self, name: str, tag: str | None) -> dict[str, t.Any] | None:
        """
        Lookup an image (by name and tag) and return the inspection results.
        """
        if not name:
            return None

        self.log(f"Find image {name}:{tag}")
        images = self._image_lookup(name, tag)
        if not images:
            # In API <= 1.20 seeing 'docker.io/<name>' as the name of images pulled from docker hub
            registry, repo_name = auth.resolve_repository_name(name)
            if registry == "docker.io":
                # If docker.io is explicitly there in name, the image
                # is not found in some cases (#41509)
                self.log(f"Check for docker.io image: {repo_name}")
                images = self._image_lookup(repo_name, tag)
                if not images and repo_name.startswith("library/"):
                    # Sometimes library/xxx images are not found
                    lookup = repo_name[len("library/") :]
                    self.log(f"Check for docker.io image: {lookup}")
                    images = self._image_lookup(lookup, tag)
                if not images:
                    # Last case for some Docker versions: if docker.io was not there,
                    # it can be that the image was not found either
                    # (https://github.com/ansible/ansible/pull/15586)
                    lookup = f"{registry}/{repo_name}"
                    self.log(f"Check for docker.io image: {lookup}")
                    images = self._image_lookup(lookup, tag)
                if not images and "/" not in repo_name:
                    # This seems to be happening with podman-docker
                    # (https://github.com/ansible-collections/community.docker/issues/291)
                    lookup = f"{registry}/library/{repo_name}"
                    self.log(f"Check for docker.io image: {lookup}")
                    images = self._image_lookup(lookup, tag)

        if len(images) > 1:
            self.fail(f"Daemon returned more than one result for {name}:{tag}")

        if len(images) == 1:
            try:
                return self.get_json("/images/{0}/json", images[0]["Id"])
            except NotFound:
                self.log(f"Image {name}:{tag} not found.")
                return None
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error inspecting image {name}:{tag} - {exc}")

        self.log(f"Image {name}:{tag} not found.")
        return None

    def find_image_by_id(
        self, image_id: str, accept_missing_image: bool = False
    ) -> dict[str, t.Any] | None:
        """
        Lookup an image (by ID) and return the inspection results.
        """
        if not image_id:
            return None

        self.log(f"Find image {image_id} (by ID)")
        try:
            return self.get_json("/images/{0}/json", image_id)
        except NotFound as exc:
            if not accept_missing_image:
                self.fail(f"Error inspecting image ID {image_id} - {exc}")
            self.log(f"Image {image_id} not found.")
            return None
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error inspecting image ID {image_id} - {exc}")

    @staticmethod
    def _compare_images(
        img1: dict[str, t.Any] | None, img2: dict[str, t.Any] | None
    ) -> bool:
        if img1 is None or img2 is None:
            return img1 == img2
        filter_keys = {"Metadata", "Identity"}
        # Since Docker 29.2 (?) the order of entries in Identity.Pull can change
        # even though the image is the same. Since we're only really interested
        # in the image's Id field, we simply skip all of Identity.
        img1_filtered = {k: v for k, v in img1.items() if k not in filter_keys}
        img2_filtered = {k: v for k, v in img2.items() if k not in filter_keys}
        return img1_filtered == img2_filtered

    def pull_image(
        self, name: str, tag: str = "latest", image_platform: str | None = None
    ) -> tuple[dict[str, t.Any] | None, bool]:
        """
        Pull an image
        """
        self.log(f"Pulling image {name}:{tag}")
        old_image = self.find_image(name, tag)
        try:
            repository, image_tag = parse_repository_tag(name)
            registry, dummy_repo_name = auth.resolve_repository_name(repository)
            params = {
                "tag": tag or image_tag or "latest",
                "fromImage": repository,
            }
            if image_platform is not None:
                params["platform"] = image_platform

            headers = {}
            header = auth.get_config_header(self, registry)
            if header:
                headers["X-Registry-Auth"] = header

            response = self._post(
                self._url("/images/create"),
                params=params,
                headers=headers,
                stream=True,
                timeout=None,
            )
            self._raise_for_status(response)
            for line in self._stream_helper(response, decode=True):
                self.log(line, pretty_print=True)
                if line.get("error"):
                    if line.get("errorDetail"):
                        error_detail = line.get("errorDetail")
                        self.fail(
                            f"Error pulling {name} - code: {error_detail.get('code')} message: {error_detail.get('message')}"
                        )
                    else:
                        self.fail(f"Error pulling {name} - {line.get('error')}")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error pulling image {name}:{tag} - {exc}")

        new_image = self.find_image(name, tag)

        return new_image, self._compare_images(old_image, new_image)


class AnsibleDockerClient(AnsibleDockerClientBase):
    def __init__(
        self,
        argument_spec: dict[str, t.Any] | None = None,
        supports_check_mode: bool = False,
        mutually_exclusive: Sequence[Sequence[str]] | None = None,
        required_together: Sequence[Sequence[str]] | None = None,
        required_if: (
            Sequence[
                tuple[str, t.Any, Sequence[str]]
                | tuple[str, t.Any, Sequence[str], bool]
            ]
            | None
        ) = None,
        required_one_of: Sequence[Sequence[str]] | None = None,
        required_by: dict[str, Sequence[str]] | None = None,
        min_docker_api_version: str | None = None,
        option_minimal_versions: dict[str, t.Any] | None = None,
        option_minimal_versions_ignore_params: Sequence[str] | None = None,
        fail_results: dict[str, t.Any] | None = None,
    ):
        # Modules can put information in here which will always be returned
        # in case client.fail() is called.
        self.fail_results = fail_results or {}

        merged_arg_spec = {}
        merged_arg_spec.update(DOCKER_COMMON_ARGS)
        if argument_spec:
            merged_arg_spec.update(argument_spec)
            self.arg_spec = merged_arg_spec

        mutually_exclusive_params: list[Sequence[str]] = []
        mutually_exclusive_params += DOCKER_MUTUALLY_EXCLUSIVE
        if mutually_exclusive:
            mutually_exclusive_params += mutually_exclusive

        required_together_params: list[Sequence[str]] = []
        required_together_params += DOCKER_REQUIRED_TOGETHER
        if required_together:
            required_together_params += required_together

        self.module = AnsibleModule(
            argument_spec=merged_arg_spec,
            supports_check_mode=supports_check_mode,
            mutually_exclusive=mutually_exclusive_params,
            required_together=required_together_params,
            required_if=required_if,
            required_one_of=required_one_of,
            required_by=required_by or {},
        )

        self.debug = self.module.params.get("debug")
        self.check_mode = self.module.check_mode

        super().__init__(min_docker_api_version=min_docker_api_version)

        if option_minimal_versions is not None:
            self._get_minimal_versions(
                option_minimal_versions, option_minimal_versions_ignore_params
            )

    def fail(self, msg: str, **kwargs: t.Any) -> t.NoReturn:
        self.fail_results.update(kwargs)
        self.module.fail_json(msg=msg, **sanitize_result(self.fail_results))

    def deprecate(
        self,
        msg: str,
        version: str | None = None,
        date: str | None = None,
        collection_name: str | None = None,
    ) -> None:
        self.module.deprecate(
            msg, version=version, date=date, collection_name=collection_name
        )

    def _get_params(self) -> dict[str, t.Any]:
        return self.module.params

    def _get_minimal_versions(
        self,
        option_minimal_versions: dict[str, t.Any],
        ignore_params: Sequence[str] | None = None,
    ) -> None:
        self.option_minimal_versions: dict[str, dict[str, t.Any]] = {}
        for option in self.module.argument_spec:
            if ignore_params is not None and option in ignore_params:
                continue
            self.option_minimal_versions[option] = {}
        self.option_minimal_versions.update(option_minimal_versions)

        for option, data in self.option_minimal_versions.items():
            # Test whether option is supported, and store result
            support_docker_api = True
            if "docker_api_version" in data:
                support_docker_api = self.docker_api_version >= LooseVersion(
                    data["docker_api_version"]
                )
            data["supported"] = support_docker_api
            # Fail if option is not supported but used
            if not data["supported"]:
                # Test whether option is specified
                if "detect_usage" in data:
                    used = data["detect_usage"](self)
                else:
                    used = self.module.params.get(option) is not None
                    if used and "default" in self.module.argument_spec[option]:
                        used = (
                            self.module.params[option]
                            != self.module.argument_spec[option]["default"]
                        )
                if used:
                    # If the option is used, compose error message.
                    if "usage_msg" in data:
                        usg = data["usage_msg"]
                    else:
                        usg = f"set {option} option"
                    if not support_docker_api:
                        msg = f"Docker API version is {self.docker_api_version_str}. Minimum version required is {data['docker_api_version']} to {usg}."
                    else:
                        # should not happen
                        msg = f"Cannot {usg} with your configuration."
                    self.fail(msg)

    def report_warnings(
        self, result: t.Any, warnings_key: Sequence[str] | None = None
    ) -> None:
        """
        Checks result of client operation for warnings, and if present, outputs them.

        warnings_key should be a list of keys used to crawl the result dictionary.
        For example, if warnings_key == ['a', 'b'], the function will consider
        result['a']['b'] if these keys exist. If the result is a non-empty string, it
        will be reported as a warning. If the result is a list, every entry will be
        reported as a warning.

        In most cases (if warnings are returned at all), warnings_key should be
        ['Warnings'] or ['Warning']. The default value (if not specified) is ['Warnings'].
        """
        if warnings_key is None:
            warnings_key = ["Warnings"]
        for key in warnings_key:
            if not isinstance(result, Mapping):
                return
            result = result.get(key)
        if isinstance(result, Sequence):
            for warning in result:
                self.module.warn(f"Docker warning: {warning}")
        elif isinstance(result, str) and result:
            self.module.warn(f"Docker warning: {result}")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Jon Meran <jonathan.meran@sonos.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: batch_compute_environment
version_added: 1.0.0
short_description: Manage AWS Batch Compute Environments
description:
  - This module allows the management of AWS Batch Compute Environments.
  - It is idempotent and supports "Check" mode.
  - Use module M(community.aws.batch_compute_environment) to manage the compute
    environment, M(community.aws.batch_job_queue) to manage job queues, M(community.aws.batch_job_definition) to manage job definitions.
  - Prior to release 5.0.0 this module was called C(community.aws.aws_batch_compute_environment).
    The usage did not change.
author:
  - Jon Meran (@jonmer85)
options:
  compute_environment_name:
    description:
      - The name for your compute environment.
      - Up to 128 letters (uppercase and lowercase), numbers, and underscores are allowed.
    required: true
    type: str
  type:
    description:
      - The type of the compute environment.
    required: true
    choices: ["MANAGED", "UNMANAGED"]
    type: str
  state:
    description:
      - Describes the desired state.
    default: "present"
    choices: ["present", "absent"]
    type: str
  compute_environment_state:
    description:
      - The state of the compute environment.
      - If the state is C(ENABLED), then the compute environment accepts jobs
        from a queue and can scale out automatically based on queues.
    default: "ENABLED"
    choices: ["ENABLED", "DISABLED"]
    type: str
  service_role:
    description:
      - The full Amazon Resource Name (ARN) of the IAM role that allows AWS Batch to make calls to other AWS
        services on your behalf.
    required: true
    type: str
  compute_resource_type:
    description:
      - The type of compute resource.
    required: true
    choices: ["EC2", "SPOT"]
    type: str
  minv_cpus:
    description:
      - The minimum number of EC2 vCPUs that an environment should maintain.
    required: true
    type: int
  maxv_cpus:
    description:
      - The maximum number of EC2 vCPUs that an environment can reach.
    required: true
    type: int
  desiredv_cpus:
    description:
      - The desired number of EC2 vCPUS in the compute environment.
    type: int
  instance_types:
    description:
      - The instance types that may be launched.
    required: true
    type: list
    elements: str
  image_id:
    description:
      - The Amazon Machine Image (AMI) ID used for instances launched in the compute environment.
    type: str
  subnets:
    description:
      - The VPC subnets into which the compute resources are launched.
    required: true
    type: list
    elements: str
  security_group_ids:
    description:
      - The EC2 security groups that are associated with instances launched in the compute environment.
    required: true
    type: list
    elements: str
  ec2_key_pair:
    description:
      - The EC2 key pair that is used for instances launched in the compute environment.
    type: str
  instance_role:
    description:
      - The Amazon ECS instance role applied to Amazon EC2 instances in a compute environment.
    required: true
    type: str
  tags:
    description:
      - Key-value pair tags to be applied to resources that are launched in the compute environment.
    type: dict
  bid_percentage:
    description:
      - The minimum percentage that a Spot Instance price must be when compared with the On-Demand price for that
        instance type before instances are launched.
      - For example, if your bid percentage is 20%, then the Spot price
        must be below 20% of the current On-Demand price for that EC2 instance.
    type: int
  spot_iam_fleet_role:
    description:
      - The Amazon Resource Name (ARN) of the Amazon EC2 Spot Fleet IAM role applied to a SPOT compute environment.
    type: str
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: My Batch Compute Environment
  community.aws.batch_compute_environment:
    compute_environment_name: computeEnvironmentName
    state: present
    region: us-east-1
    compute_environment_state: ENABLED
    type: MANAGED
    compute_resource_type: EC2
    minv_cpus: 0
    maxv_cpus: 2
    desiredv_cpus: 1
    instance_types:
      - optimal
    subnets:
      - my-subnet1
      - my-subnet2
    security_group_ids:
      - my-sg1
      - my-sg2
    instance_role: arn:aws:iam::<account>:instance-profile/<role>
    tags:
      tag1: value1
      tag2: value2
    service_role: arn:aws:iam::<account>:role/service-role/<role>
  register: aws_batch_compute_environment_action

- name: show results
  ansible.builtin.debug:
    var: aws_batch_compute_environment_action
"""

RETURN = r"""
---
output:
  description: "returns what action was taken, whether something was changed, invocation and response"
  returned: always
  sample:
    batch_compute_environment_action: none
    changed: false
    invocation:
      module_args:
        access_key: ~
        secret_key: ~
        bid_percentage: ~
        compute_environment_name: <name>
        compute_environment_state: ENABLED
        compute_resource_type: EC2
        desiredv_cpus: 0
        ec2_key_pair: ~
        endpoint_url: ~
        image_id: ~
        instance_role: "arn:aws:iam::..."
        instance_types:
          - optimal
        maxv_cpus: 8
        minv_cpus: 0
        profile: ~
        region: us-east-1
        security_group_ids:
          - "*******"
        security_token: ~
        service_role: "arn:aws:iam::...."
        spot_iam_fleet_role: ~
        state: present
        subnets:
          - "******"
        tags:
          Environment: <name>
          Name: <name>
        type: MANAGED
        validate_certs: true
    response:
      computeEnvironmentArn: "arn:aws:batch:...."
      computeEnvironmentName: <name>
      computeResources:
        desiredvCpus: 0
        instanceRole: "arn:aws:iam::..."
        instanceTypes:
          - optimal
        maxvCpus: 8
        minvCpus: 0
        securityGroupIds:
          - "******"
        subnets:
          - "*******"
        tags:
          Environment: <name>
          Name: <name>
        type: EC2
      ecsClusterArn: "arn:aws:ecs:....."
      serviceRole: "arn:aws:iam::..."
      state: ENABLED
      status: VALID
      statusReason: "ComputeEnvironment Healthy"
      type: MANAGED
  type: dict
"""

import re

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict
from ansible.module_utils.common.dict_transformations import snake_dict_to_camel_dict

from ansible_collections.amazon.aws.plugins.module_utils.arn import validate_aws_arn

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule

# ---------------------------------------------------------------------------------------------------
#
#   Helper Functions & classes
#
# ---------------------------------------------------------------------------------------------------


def set_api_params(module, module_params):
    """
    Sets module parameters to those expected by the boto3 API.

    :param module:
    :param module_params:
    :return:
    """
    api_params = dict((k, v) for k, v in dict(module.params).items() if k in module_params and v is not None)
    return snake_dict_to_camel_dict(api_params)


def validate_params(module):
    """
    Performs basic parameter validation.

    :param module:
    :return:
    """

    compute_environment_name = module.params["compute_environment_name"]

    # validate compute environment name
    if not re.search(r"^[\w\_:]+$", compute_environment_name):
        module.fail_json(
            msg=f"Function compute_environment_name {compute_environment_name} is invalid. Names must contain only alphanumeric characters and underscores."
        )
    if not validate_aws_arn(compute_environment_name, service="batch"):
        if len(compute_environment_name) > 128:
            module.fail_json(msg=f'compute_environment_name "{compute_environment_name}" exceeds 128 character limit')

    return


# ---------------------------------------------------------------------------------------------------
#
#   Batch Compute Environment functions
#
# ---------------------------------------------------------------------------------------------------


def get_current_compute_environment(module, client):
    try:
        environments = client.describe_compute_environments(
            computeEnvironments=[module.params["compute_environment_name"]]
        )
        if len(environments["computeEnvironments"]) > 0:
            return environments["computeEnvironments"][0]
        else:
            return None
    except ClientError:
        return None


def create_compute_environment(module, client):
    """
    Adds a Batch compute environment

    :param module:
    :param client:
    :return:
    """

    changed = False

    # set API parameters
    params = ("compute_environment_name", "type", "service_role")
    api_params = set_api_params(module, params)

    if module.params["compute_environment_state"] is not None:
        api_params["state"] = module.params["compute_environment_state"]

    compute_resources_param_list = (
        "minv_cpus",
        "maxv_cpus",
        "desiredv_cpus",
        "instance_types",
        "image_id",
        "subnets",
        "security_group_ids",
        "ec2_key_pair",
        "instance_role",
        "tags",
        "bid_percentage",
        "spot_iam_fleet_role",
    )
    compute_resources_params = set_api_params(module, compute_resources_param_list)

    if module.params["compute_resource_type"] is not None:
        compute_resources_params["type"] = module.params["compute_resource_type"]

    # if module.params['minv_cpus'] is not None:
    #     compute_resources_params['minvCpus'] = module.params['minv_cpus']

    api_params["computeResources"] = compute_resources_params

    try:
        if not module.check_mode:
            client.create_compute_environment(**api_params)
        changed = True
    except (ClientError, BotoCoreError) as e:
        module.fail_json_aws(e, msg="Error creating compute environment")

    return changed


def remove_compute_environment(module, client):
    """
    Remove a Batch compute environment

    :param module:
    :param client:
    :return:
    """

    changed = False

    # set API parameters
    api_params = {"computeEnvironment": module.params["compute_environment_name"]}

    try:
        if not module.check_mode:
            client.delete_compute_environment(**api_params)
        changed = True
    except (ClientError, BotoCoreError) as e:
        module.fail_json_aws(e, msg="Error removing compute environment")
    return changed


def manage_state(module, client):
    changed = False
    current_state = "absent"
    state = module.params["state"]
    compute_environment_state = module.params["compute_environment_state"]
    compute_environment_name = module.params["compute_environment_name"]
    service_role = module.params["service_role"]
    minv_cpus = module.params["minv_cpus"]
    maxv_cpus = module.params["maxv_cpus"]
    desiredv_cpus = module.params["desiredv_cpus"]
    action_taken = "none"
    update_env_response = ""

    check_mode = module.check_mode

    # check if the compute environment exists
    current_compute_environment = get_current_compute_environment(module, client)
    response = current_compute_environment
    if current_compute_environment:
        current_state = "present"

    if state == "present":
        if current_state == "present":
            updates = False
            # Update Batch Compute Environment configuration
            compute_kwargs = {"computeEnvironment": compute_environment_name}

            # Update configuration if needed
            compute_resources = {}
            if compute_environment_state and current_compute_environment["state"] != compute_environment_state:
                compute_kwargs.update({"state": compute_environment_state})
                updates = True
            if service_role and current_compute_environment["serviceRole"] != service_role:
                compute_kwargs.update({"serviceRole": service_role})
                updates = True
            if minv_cpus is not None and current_compute_environment["computeResources"]["minvCpus"] != minv_cpus:
                compute_resources["minvCpus"] = minv_cpus
            if maxv_cpus is not None and current_compute_environment["computeResources"]["maxvCpus"] != maxv_cpus:
                compute_resources["maxvCpus"] = maxv_cpus
            if (
                desiredv_cpus is not None
                and current_compute_environment["computeResources"]["desiredvCpus"] != desiredv_cpus
            ):
                compute_resources["desiredvCpus"] = desiredv_cpus
            if len(compute_resources) > 0:
                compute_kwargs["computeResources"] = compute_resources
                updates = True
            if updates:
                try:
                    if not check_mode:
                        update_env_response = client.update_compute_environment(**compute_kwargs)
                    if not update_env_response:
                        module.fail_json(msg="Unable to get compute environment information after creating")
                    changed = True
                    action_taken = "updated"
                except (BotoCoreError, ClientError) as e:
                    module.fail_json_aws(e, msg="Unable to update environment.")

        else:
            # Create Batch Compute Environment
            changed = create_compute_environment(module, client)
            # Describe compute environment
            action_taken = "added"
        response = get_current_compute_environment(module, client)
        if not response:
            module.fail_json(msg="Unable to get compute environment information after creating")
    else:
        if current_state == "present":
            # remove the compute environment
            changed = remove_compute_environment(module, client)
            action_taken = "deleted"
    return dict(changed=changed, batch_compute_environment_action=action_taken, response=response)


# ---------------------------------------------------------------------------------------------------
#
#   MAIN
#
# ---------------------------------------------------------------------------------------------------


def main():
    """
    Main entry point.

    :return dict: changed, batch_compute_environment_action, response
    """

    argument_spec = dict(
        state=dict(default="present", choices=["present", "absent"]),
        compute_environment_name=dict(required=True),
        type=dict(required=True, choices=["MANAGED", "UNMANAGED"]),
        compute_environment_state=dict(required=False, default="ENABLED", choices=["ENABLED", "DISABLED"]),
        service_role=dict(required=True),
        compute_resource_type=dict(required=True, choices=["EC2", "SPOT"]),
        minv_cpus=dict(type="int", required=True),
        maxv_cpus=dict(type="int", required=True),
        desiredv_cpus=dict(type="int"),
        instance_types=dict(type="list", required=True, elements="str"),
        image_id=dict(),
        subnets=dict(type="list", required=True, elements="str"),
        security_group_ids=dict(type="list", required=True, elements="str"),
        ec2_key_pair=dict(no_log=False),
        instance_role=dict(required=True),
        tags=dict(type="dict"),
        bid_percentage=dict(type="int"),
        spot_iam_fleet_role=dict(),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    client = module.client("batch")

    validate_params(module)

    results = manage_state(module, client)

    module.exit_json(**camel_dict_to_snake_dict(results, ignore_list=["Tags"]))


if __name__ == "__main__":
    main()
